@@ -41,7 +41,14 @@
 
 #define HERE printf("at %s\r\n", LOC);
 
-//=============================================================================
+//----------------------------------------------------------------------------
+extern int64_t get_main();
+int64_t
+get_main()
+{
+   return 0;
+}
+//============================================================================
 static int init_done = 0;
 
 struct value_callback_context
