@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2018  Dr. Jürgen Sauermann
+    Copyright © 2008-2023  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -18,10 +18,13 @@
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
+/** @file
+*/
+
 #include "CharCell.hh"
 #include "Workspace.hh"
 
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 CellType
 CharCell::get_cell_subtype() const
 {
@@ -63,7 +66,7 @@ CharCell::get_cell_subtype() const
    return CellType(CT_CHAR | CTS_X32 | CTS_S32 | CTS_U32 |
                    CTS_X64 | CTS_S64 | CTS_U64);
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 bool
 CharCell::greater(const Cell & other) const
 {
@@ -80,29 +83,34 @@ const Unicode other_val = other.get_char_value();
 
    return this_val > other_val;
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 bool
 CharCell::equal(const Cell & other, double qct) const
 {
    if (!other.is_character_cell())   return false;
    return value.aval == other.get_char_value();
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 Comp_result
 CharCell::compare(const Cell & other) const
 {
-   if (other.get_char_value() == value.aval)  return COMP_EQ;
-   return (value.aval < other.get_char_value()) ? COMP_LT : COMP_GT;
+   if (other.is_character_cell())
+      {
+        if (other.get_char_value() == value.aval)  return COMP_EQ;
+        return (value.aval < other.get_char_value()) ? COMP_LT : COMP_GT;
+      }
+
+   return COMP_LT;   // char < everything else
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 bool
 CharCell::is_example_field() const
 {
-   if (value.aval == UNI_ASCII_COMMA)       return true;
-   if (value.aval == UNI_ASCII_FULLSTOP)    return true;
-   return value.aval >= UNI_ASCII_0 && value.aval <= UNI_ASCII_9;
+   if (value.aval == UNI_COMMA)       return true;
+   if (value.aval == UNI_FULLSTOP)    return true;
+   return value.aval >= UNI_0 && value.aval <= UNI_9;
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 PrintBuffer
 CharCell::character_representation(const PrintContext & pctx) const
 {
@@ -112,7 +120,7 @@ ColInfo info;
 
 PrintStyle style = pctx.get_style();
 Unicode uni = get_char_value();
-   if ((style & PST_PRETTY) && uni < UNI_ASCII_SPACE)
+   if ((style & PST_PRETTY) && uni < UNI_SPACE)
       uni = Unicode(uni + 0x2400);
 
    if (style == PR_APL_FUN)
@@ -149,14 +157,14 @@ Unicode uni = get_char_value();
 
    return PrintBuffer(ucs, info);
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 int
 CharCell::get_byte_value() const
 {
    if (value.aval < -128)
       {
         char cc[20];
-        snprintf(cc, sizeof(cc), "' = U+%4.4X", value.aval & 0x0FFFF);
+        SPRINTF(cc, "' = U+%4.4X", value.aval & 0x0FFFF);
         MORE_ERROR() << "Unicode character '" << value.aval << cc
                      << " is too small (< U+FF80) for a byte value";
         DOMAIN_ERROR;
@@ -165,7 +173,7 @@ CharCell::get_byte_value() const
    if (value.aval > 255)
       {
         char cc[20];
-        snprintf(cc, sizeof(cc), "' = U+%4.4X", value.aval & 0x0FFFF);
+        SPRINTF(cc, "' = U+%4.4X", value.aval & 0x0FFFF);
         MORE_ERROR() << "Unicode character '" << value.aval << cc
                      << " is too large (> U+00FF) for a byte value";
         DOMAIN_ERROR;
@@ -173,7 +181,7 @@ CharCell::get_byte_value() const
 
    return value.aval;
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 int
 CharCell::CDR_size() const
 {
@@ -184,55 +192,58 @@ const Unicode uni = get_char_value();
    if (uni >= 256)   return 4;
    return 1;
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 ErrorCode
 CharCell::bif_not_bitwise(Cell * Z) const
 {
-   return zv(Z, Unicode(get_char_value() ^ 0xFFFFFFFF));
+   return zU(Z, Unicode(get_char_value() ^ 0xFFFFFFFF));
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 ErrorCode
 CharCell::bif_and_bitwise(Cell * Z, const Cell * A) const
 {
    if (A->is_character_cell())
-     return zv(Z, Unicode(value.aval & A->get_char_value()));
-   else if (A->is_numeric())
-      return zv(Z, Unicode(value.aval & A->get_int_value()));
-   else
-      return E_DOMAIN_ERROR;
+     return zU(Z, Unicode(value.aval & A->get_char_value()));
+
+   if (A->is_numeric())
+      return zU(Z, Unicode(value.aval & A->get_int_value()));
+
+   return E_DOMAIN_ERROR;
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 ErrorCode
 CharCell::bif_or_bitwise(Cell * Z, const Cell * A) const
 {
    if (A->is_character_cell())
-     return zv(Z, Unicode(value.aval | A->get_char_value()));
-   else if (A->is_numeric())
-      return zv(Z, Unicode(value.aval | A->get_int_value()));
-   else
-      return E_DOMAIN_ERROR;
+     return zU(Z, Unicode(value.aval | A->get_char_value()));
+
+   if (A->is_numeric())
+      return zU(Z, Unicode(value.aval | A->get_int_value()));
+
+   return E_DOMAIN_ERROR;
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 ErrorCode
 CharCell::bif_equal_bitwise(Cell * Z, const Cell * A) const
 {
    if (A->is_character_cell())
-     return zv(Z, Unicode(0xFFFFFFFF & (value.aval ^ A->get_char_value())));
-   else if (A->is_numeric())
-      return zv(Z, Unicode(0xFFFFFFFF & (value.aval ^ A->get_int_value())));
-   else
-      return E_DOMAIN_ERROR;
+     return zU(Z, Unicode(0xFFFFFFFF & (value.aval ^ A->get_char_value())));
+
+   if (A->is_numeric())
+      return zU(Z, Unicode(0xFFFFFFFF & (value.aval ^ A->get_int_value())));
+   return E_DOMAIN_ERROR;
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 ErrorCode
 CharCell::bif_not_equal_bitwise(Cell * Z, const Cell * A) const
 {
    if (A->is_character_cell())
-     return zv(Z, Unicode(value.aval ^ A->get_char_value()));
-   else if (A->is_numeric())
-      return zv(Z, Unicode(value.aval ^ A->get_int_value()));
-   else
-      return E_DOMAIN_ERROR;
+     return zU(Z, Unicode(value.aval ^ A->get_char_value()));
+
+   if (A->is_numeric())
+      return zU(Z, Unicode(value.aval ^ A->get_int_value()));
+
+   return E_DOMAIN_ERROR;
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 
