@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
+    Copyright © 2008-2023  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -18,6 +18,9 @@
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
+/** @file
+*/
+
 #include "Common.hh"
 #include "Error.hh"
 #include "Output.hh"
@@ -27,7 +30,9 @@
 #include "UserFunction.hh"
 #include "Workspace.hh"
 
-//-----------------------------------------------------------------------------
+#include "Workspace.icc"
+
+//----------------------------------------------------------------------------
 Error::Error(ErrorCode ec, const char * loc)
    : error_code(ec),
      throw_loc(loc),
@@ -37,14 +42,25 @@ Error::Error(ErrorCode ec, const char * loc)
      right_caret(-1),
      print_loc(0)
 {
-const char more = Workspace::more_error().size() ? '+' : 0;
-   snprintf(error_message_1, sizeof(error_message_1), "%s%c",
-            error_name(error_code), more);
+const bool have_more = Workspace::more_error().size();
+   if (have_more && Workspace::more_error().back() != UNI_PLUS)
+      {
+        SPRINTF(error_message_1, "%s%c", error_name(error_code), UNI_PLUS);
+      }
+   else
+      {
+        SPRINTF(error_message_1, "%s", error_name(error_code));
+      }
+
+   if (have_more && Command::auto_MORE)
+      {
+        CERR << Workspace::more_error() << endl;
+      }
 
    *symbol_name = 0;
    *error_message_2 = 0;
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 void
 Error::print(ostream & out, const char * loc) const
 {
@@ -77,7 +93,7 @@ Error::print(ostream & out, const char * loc) const
         else
            {
              out << error_name(error_code);
-             if (Workspace::more_error().size())   out << UNI_ASCII_PLUS;
+             if (Workspace::more_error().back() != UNI_PLUS)   out << UNI_PLUS;
              out << endl;
            }
 
@@ -94,7 +110,7 @@ Error::print(ostream & out, const char * loc) const
                                                                   << endl;
       }
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 const char *
 Error::error_name(ErrorCode err)
 {
@@ -109,7 +125,7 @@ Error::error_name(ErrorCode err)
 
    return "Unknown Error";
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 bool
 Error::is_known() const
 {
@@ -122,7 +138,36 @@ Error::is_known() const
 
    return false;
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
+void
+Error::set_error_line_1(const char * msg_1)
+{
+   strncpy(error_message_1, msg_1, sizeof(error_message_1));
+   error_message_1[sizeof(error_message_1) - 1] = 0;
+}
+//----------------------------------------------------------------------------
+void
+Error::add_MORE_indicator(bool have_more)
+{
+   if (have_more)   // )MORE info available
+      {
+        const size_t len = strlen(error_message_1);
+        if (error_message_1[len - 1] != UNI_PLUS &&
+            len < sizeof(error_message_1) - 1)
+           {
+             error_message_1[len]     = '+';
+             error_message_1[len + 1] = 0;
+           }
+      }
+}
+//----------------------------------------------------------------------------
+void
+Error::set_error_line_2(const char * msg_2)
+{
+   strncpy(error_message_2, msg_2, sizeof(error_message_2));
+   error_message_2[sizeof(error_message_2) - 1] = 0;
+}
+//----------------------------------------------------------------------------
 void
 Error::set_error_line_2(const UCS_string & ucs, int lcaret, int rcaret)
 {
@@ -132,7 +177,7 @@ UTF8_string utf(ucs);
    left_caret = lcaret;
    right_caret = rcaret;
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 bool
 Error::is_syntax_or_value_error() const
 {
@@ -140,26 +185,26 @@ Error::is_syntax_or_value_error() const
    if (error_major(error_code) == 3)   return true;   // VALUE ERROR
    return false;
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 UCS_string
 Error::get_error_line_3() const
 {
-   if (error_code == E_NO_ERROR)   return UCS_string();
+   if (error_code == E_NO_ERROR)   return UCS_string();   // no error
 
-   if (left_caret < 0)   return UCS_string();
+   if (left_caret < 0)   return UCS_string();             // no ^ position
 
 UCS_string ret;
-   if (left_caret > 0)   ret.append(UCS_string(left_caret, UNI_ASCII_SPACE));
-   ret.append(UNI_ASCII_CIRCUMFLEX);
+   if (left_caret > 0)   ret.append(UCS_string(left_caret, UNI_SPACE));
+   ret.append(UNI_CIRCUMFLEX);
 
 const int diff = right_caret - left_caret;
    if (diff <= 0)   return ret;
-   if (diff > 1)   ret.append(UCS_string(diff - 1, UNI_ASCII_SPACE));
-   ret.append(UNI_ASCII_CIRCUMFLEX);
+   if (diff > 1)   ret.append(UCS_string(diff - 1, UNI_SPACE));
+   ret.append(UNI_CIRCUMFLEX);
 
    return ret;
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 void
 Error::print_em(ostream & out, const char * loc)
 {
@@ -173,18 +218,18 @@ Error::print_em(ostream & out, const char * loc)
       }
 
    print_loc = loc;
-   if (*get_error_line_1())   out << get_error_line_1() << endl;
+   out << get_error_line_1() << endl;
 
    out << get_error_line_2() << endl
        << get_error_line_3() << endl;
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 void
 throw_apl_error(ErrorCode code, const char * loc)
 {
    ADD_EVENT(0, VHE_Error, code, loc);
 
-StateIndicator * si = Workspace::SI_top();
+StateIndicator * si = Workspace::SI_top();   // the current )SI entry
 
    Log(LOG_error_throw)
       {
@@ -195,15 +240,14 @@ StateIndicator * si = Workspace::SI_top();
 
    Log(LOG_verbose_error)
       {
-        if (!(si && si->get_safe_execution()))
-           Backtrace::show(__FILE__, __LINE__);
+        if (!(si && si->get_safe_execution_count()))   BACKTRACE
       }
 
    // maybe map error to DOMAIN ERROR.
    //
    if (si)
       {
-        const UserFunction * ufun = si->get_executable()->get_ufun();
+        const UserFunction * ufun = si->get_executable()->get_exec_ufun();
         if (ufun && ufun->get_exec_properties()[3])   code = E_DOMAIN_ERROR;
       }
 
@@ -213,7 +257,7 @@ Error error(code, loc);
 Error & eref = error;
    throw eref;
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 void
 Error::throw_parse_error(ErrorCode code, const char * par_loc, const char *loc)
 {
@@ -221,7 +265,7 @@ Error::throw_parse_error(ErrorCode code, const char * par_loc, const char *loc)
       CERR << endl
            << "throwing " << Error::error_name(code) << " at " << loc << endl;
 
-   Log(LOG_verbose_error)   Backtrace::show(__FILE__, __LINE__);
+   Log(LOG_verbose_error)   BACKTRACE
 
    MORE_ERROR() << Error::error_name(code);
 
@@ -234,119 +278,137 @@ Error error(code, loc);
 Error & eref = error;
    throw eref;
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 void
 Error::throw_symbol_error(const UCS_string & sym_name, const char * loc)
 {
    Log(LOG_error_throw)   
       {   
         CERR << "throwing VALUE ERROR at " << loc;
-        if (sym_name.size())   CERR << " (symbol is " << sym_name << ")"; 
+        if (sym_name.size())   CERR << " (symbol is: '" << sym_name << "')"; 
         CERR << endl;
       }
 
-   Log(LOG_verbose_error)     Backtrace::show(__FILE__, __LINE__);
+   if (Workspace::more_error().size() == 0)   // no )MORE info provided
+      {
+        MORE_ERROR() << "Offending symbol: " << sym_name;
+      }
 
-Error error(E_VALUE_ERROR, loc);
+   Log(LOG_verbose_error)     BACKTRACE
+
+Error err(E_VALUE_ERROR, loc);
 UTF8_string sym_name_utf(sym_name);
-   snprintf(error.symbol_name, sizeof(error.symbol_name), "%s",
-            sym_name_utf.c_str());
-   if (StateIndicator * si = Workspace::SI_top())   error.update_error_info(si);
+   SPRINTF(err.symbol_name, "%s", sym_name_utf.c_str());
 
-Error & eref = error;
+   if (StateIndicator * si = Workspace::SI_top())   // )SI not empty
+      err.update_error_info(si);
+
+Error & eref = err;
    throw eref;
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 void
 Error::throw_define_error(const UCS_string & fun_name, const UCS_string & cmd,
                           const char * loc)
 {
+   // this error is thrown if the ∇-editor detects an invalid function header,
+   // for example: "∇FOO 2"
+
    Log(LOG_error_throw)   
       {   
         CERR << "throwing DEFN ERROR at " << loc
              << " (function is " << fun_name << ")" << endl;
       }
 
-   Log(LOG_verbose_error)     Backtrace::show(__FILE__, __LINE__);
+   Log(LOG_verbose_error)     BACKTRACE
 
-Error error(E_DEFN_ERROR, loc);
-Error & eref = error;
+Error err(E_DEFN_ERROR, loc);
+Error & eref = err;
 UTF8_string fun_name_utf(fun_name);
-   snprintf(error.symbol_name, sizeof(error.symbol_name), "%s",
-            fun_name_utf.c_str());
+   SPRINTF(err.symbol_name, "%s", fun_name_utf.c_str());
 
 UTF8_string cmd_utf(cmd);   // cmd is something like ∇FUN[⎕]∇
-   snprintf(error.error_message_2, sizeof(error.error_message_2),
-            " aaa  %s", cmd_utf.c_str());
-   eref.left_caret = 5 + cmd.size();
+   SPRINTF(err.error_message_2, "PROMPT%s", cmd_utf.c_str());
+   eref.left_caret = 5 + cmd.size();   // last Unicode of error_message_2
    if (Workspace::SI_top())   *Workspace::get_error() = eref;
    throw eref;
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 void
 Error::update_error_info(StateIndicator * si)
 {
-bool locked = false;
-const UserFunction * ufun = si->get_executable()->get_ufun();
+   /*
+      construct lines 2 and 3 of the standard 3-line APL error message.
+      Line 1 was already constructed in Error::Error(ErrorCode ec...).
+      For eample: Let
 
-   // prepare second error line (failed statement)
+            ∇FOO
+      [1] 1 2 3 ◊ Q←Q++
+      [2] ∇
+
+      Then:
+
+            FOO                   ⍝ APL code producing the error
+      1 2 3                       ⍝ first (correct) statement
+      VALUE ERROR                 ⍝ error line 1
+      FOO[1]  Q←Q++               ⍝ error line 2: failed statement
+                ^                 ⍝ error line 3: caret line (error range)
+
+       There is no error_message_3 in class Error; the third error line
+       is constructed from left_caret and right_caret when needed.
+
+       lrm: "The left caret indicates how far execution of the expression
+             progressed before the suspension occurred.
+             The right caret indicates the likely point of the error. (On
+             occasion, the two carets overlap so that only one is displayed.)"
+    */
+
+   set_error_line_2("      ");   // the APL prompt
+   set_right_caret(-1);
+   set_left_caret(6);            // first char after the APL prompt
+
+   // prepare the second error line (= display of the failed statement)
    //
-   if (ufun)
+   if (const UserFunction * ufun = si->get_executable()->get_exec_ufun())
       {
+        // ufun->print_line_PCs(LOC);
         if (get_show_locked() || ufun->get_exec_properties()[1])
            {
-             locked = true;
-             set_error_line_2("      ");
-             set_left_caret(6);
+             set_error_line_2("      ");   // the APL prompt
+             set_left_caret(6);            // first char after the APL prompt
+             ufun->set_locked_error_info(*this);
+             goto out;   // maybe print
            }
-        else
-           {
-             UCS_string ucs(ufun->get_name_and_line(si->get_PC()));
-             ucs.append(UNI_ASCII_SPACE);
-             ucs.append(UNI_ASCII_SPACE);
-             UTF8_string utf(ucs);
-             set_error_line_2(utf.c_str());
-             set_left_caret(ucs.size());
-           }
-      }
-   else
-      {
-        set_error_line_2("      ");
-        set_left_caret(6);
+
+        UCS_string ucs(ufun->get_name_and_line(si->get_PC()));
+        ucs.append(UNI_SPACE);
+        ucs.append(UNI_SPACE);
+        UTF8_string utf(ucs);
+        set_error_line_2(utf.c_str());
+        set_left_caret(ucs.size());
       }
 
-   // prepare third line (carets)
-   //
-   set_right_caret(-1);
+   {
+     const Prefix & prefix = si->get_prefix();
+     const Function_PC from = prefix.get_range_low();
+     const Function_PC to   = prefix.get_range_high();
+     const Function_PC2 error_range(from, to);
 
-   if (locked)
-      {
-        si->get_executable()->get_ufun()->set_locked_error_info(*this);
-      }
-   else
-      {
-        const Function_PC from = si->get_prefix().get_range_low();
-        Function_PC to   = si->get_prefix().get_range_high();
-        const Function_PC2 error_range(from, to);
-        si->get_executable()->set_error_info(*this, error_range);
-      }
+     si->get_executable()->set_error_info(*this, error_range);
+   }
 
    // print error, unless we are in safe execution mode.
    //
-   {
-     bool print_error = true;
-     for (const StateIndicator * si1 = si; si1; si1 = si1->get_parent())
-         {
-           if (si1->get_safe_execution())
-              {
-                print_error = false;
-                break;
-              }
-         }
-     if (print_error)   print_em(UERR, LOC);
-   }
-
+out:
    StateIndicator::get_error(si) = *this;
+
+   // )SI entries below a ⎕ES entry must not print anything but simply return
+   for (const StateIndicator * si1 = si; si1; si1 = si1->get_parent())
+       {
+         if (si1->get_safe_execution_count()) return;
+       }
+
+   print_em(UERR, LOC);
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 
