@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2016  Dr. Jürgen Sauermann
+    Copyright © 2008-2023  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -18,12 +18,12 @@
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
+/** @file
+*/
+
 #include <limits.h>
 #include <string.h>
 #include <sys/time.h>
-#include <unistd.h>
-
-#include "../config.h"
 
 #include "Common.hh"
 #include "InputFile.hh"
@@ -33,11 +33,12 @@
 #include "UserFunction.hh"
 #include "UserPreferences.hh"
 
-Simple_string<InputFile, false> InputFile::files_todo;
-Simple_string<InputFile, false> InputFile::files_orig;
+vector<InputFile> InputFile::files_todo;
+vector<InputFile> InputFile::files_orig;
 int InputFile::stdin_line_no = 1;
+int64_t InputFile::next_file_seq = 0;
 
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 void
 InputFile::open_current_file()
 {
@@ -53,7 +54,7 @@ InputFile::open_current_file()
         files_todo[0].line_no = 0;
       }
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 void
 InputFile::close_current_file()
 {
@@ -76,7 +77,7 @@ InputFile::close_current_file()
            }
       }
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 void
 InputFile::randomize_files()
 {
@@ -98,7 +99,7 @@ InputFile::randomize_files()
            }
       }
 
-   for (int done = 0; done < 4*files_todo.size();)
+   for (size_t done = 0; done < 4*files_todo.size();)
        {
          const int n1 = random() % files_todo.size();
          const int n2 = random() % files_todo.size();
@@ -126,14 +127,15 @@ InputFile::randomize_files()
          ++done;
        }
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 bool
 InputFile::echo_current_file()
 {
    if (files_todo.size()) return files_todo[0].echo;
-   return uprefs.echo_CIN || !uprefs.do_not_echo;
+   return UserPreferences::uprefs.echo_CIN ||
+          ! UserPreferences::uprefs.do_not_echo;
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 bool
 InputFile::check_filter(const UTF8_string & line)
 {
@@ -192,7 +194,7 @@ UCS_string ucs_line(line);
            {
              if (ucs_line[u] != UNI_LEFT_ARROW)   continue;   // not ←
 
-             ucs_line.shrink(u);
+             ucs_line.resize(u);
              ucs_line.remove_leading_and_trailing_whitespaces();
              loop(n, object_filter.size())
                  {
@@ -207,5 +209,5 @@ UCS_string ucs_line(line);
 
    return ret;
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 
