@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
+    Copyright © 2008-2023  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -18,6 +18,9 @@
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
+/** @file
+*/
+
 #include <stdio.h>
 #include <math.h>
 
@@ -29,6 +32,9 @@
 #include "IntCell.hh"
 #include "Workspace.hh"
 
+const IntCell IntCell::boolean_FALSE(0);
+const IntCell IntCell::boolean_TRUE(1);
+
 /*
    general comment: for operations between an IntCell and a "higher" numeric
    Cell type (i.e. FloatCell or ComplexCell), we try to call the corresponding
@@ -39,7 +45,7 @@
    INT ≤ FLOAT  --> FLOAT ≥ INT   (≤ is the opposite of ≥)
 
  */
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 CellType
 IntCell::get_cell_subtype() const
 {
@@ -106,17 +112,17 @@ IntCell::get_cell_subtype() const
 
    return CellType(CT_INT | CTS_U64);
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 bool
 IntCell::equal(const Cell & other, double qct) const
 {
    if (other.is_integer_cell())    return value.ival == other.get_int_value();
-   if (!other.is_numeric())        return false;
-   return other.equal(*this, qct);
+   if (other.is_numeric())         return other.equal(*this, qct);
+   return false;
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 // monadic built-in functions...
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 
 const APL_Integer int_factorials[] =
 {
@@ -333,78 +339,72 @@ const APL_Integer iv = value.ival;
    // 64-bit signed integer for N <= 20
    // 64-bit float for N <= 170
    //
-   
    if (iv < 0)     return E_DOMAIN_ERROR;
 
    if (iv < int_fact_max)   // integer result that fits into a 64-bit int
       {
-        new (Z) IntCell(int_factorials[iv]);
-        return E_NO_ERROR;
+        return IntCell::zI(Z, int_factorials[iv]);
       }
 
    if (iv < float_fact_max)   // float result that fits into a 64 bit double
       {
-        new (Z) FloatCell(float_factorials[iv]);
-        return E_NO_ERROR;
+        return FloatCell::zF(Z, float_factorials[iv]);
       }
 
    return E_DOMAIN_ERROR;
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 ErrorCode
 IntCell::bif_conjugate(Cell * Z) const
 {
-   new (Z) IntCell(value.ival);
-   return E_NO_ERROR;
+   return IntCell::zI(Z, value.ival);
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 ErrorCode
 IntCell::bif_negative(Cell * Z) const
 {
    if (uint64_t(value.ival) == 0x8000000000000000LL)   // integer overflow
-      new (Z) FloatCell(- value.ival);
-   else
-      new (Z) IntCell(- value.ival);
-   return E_NO_ERROR;
+      return IntCell::zI(Z, - value.ival);
+   return  IntCell::zI(Z, - value.ival);
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 ErrorCode
 IntCell::bif_direction(Cell * Z) const
 {
-   if      (value.ival > 0)   new (Z) IntCell( 1);
-   else if (value.ival < 0)   new (Z) IntCell(-1);
-   else                       new (Z) IntCell( 0);
-   return E_NO_ERROR;
+APL_Integer result = 0;
+   if      (value.ival > 0)   result =  1;
+   else if (value.ival < 0)   result = -1;
+   return IntCell::zI(Z, result);
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 ErrorCode
 IntCell::bif_magnitude(Cell * Z) const
 {
-   if (value.ival >= 0)   new (Z) IntCell(value.ival);
+   if (value.ival >= 0)   IntCell::zI(Z, value.ival);
    else                   bif_negative(Z);
    return E_NO_ERROR;
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 ErrorCode
 IntCell::bif_reciprocal(Cell * Z) const
 {
    switch(value.ival)
       {
         case  0: return E_DOMAIN_ERROR;
-        case  1: return IntCell::zv(Z,  1);
-        case -1: return IntCell::zv(Z, -1);
+        case  1: return IntCell::z1(Z);
+        case -1: return IntCell::z_1(Z);
       }
 
-#ifdef RATIONAL_NUMBERS_WANTED
-   if (value.ival < 0)
-      return FloatCell::zv(Z, -1, -value.ival);
+#ifdef cfg_RATIONAL_NUMBERS_WANTED
+   if (value.ival < 0)   // make denominator positive
+      return FloatCell::zR(Z, -1, -value.ival);   // 1 ÷ X = ¯1 ÷ -X
    else
-      return FloatCell::zv(Z, 1, value.ival);
+      return FloatCell::zR(Z, 1, value.ival);     // 1 ÷ X
 #endif
 
-   return FloatCell::zv(Z, 1.0/value.ival);
+   return FloatCell::zF(Z, 1.0/value.ival);
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 ErrorCode
 IntCell::bif_roll(Cell * Z) const
 {
@@ -413,45 +413,41 @@ const APL_Integer set_size = value.ival;
    if (set_size <= 0)   return E_DOMAIN_ERROR;
 
 const uint64_t rnd = Workspace::get_RL(set_size);
-   new (Z) IntCell(qio + (rnd % set_size));
-   return E_NO_ERROR;
+   return IntCell::zI(Z, qio + (rnd % set_size));
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 ErrorCode
 IntCell::bif_pi_times(Cell * Z) const
 {
-   new (Z) FloatCell(M_PI * value.ival);
-   return E_NO_ERROR;
+   return FloatCell::zF(Z, M_PI * value.ival);
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 ErrorCode
 IntCell::bif_pi_times_inverse(Cell * Z) const
 {
-   new (Z) FloatCell(value.ival / M_PI);
-   return E_NO_ERROR;
+   return FloatCell::zF(Z, value.ival / M_PI);
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 ErrorCode
 IntCell::bif_ceiling(Cell * Z) const
 {
-   return zv(Z, value.ival);
+   return IntCell::zI(Z, value.ival);
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 ErrorCode
 IntCell::bif_floor(Cell * Z) const
 {
-   return zv(Z, value.ival);
+   return IntCell::zI(Z, value.ival);
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 ErrorCode
 IntCell::bif_exponential(Cell * Z) const
 {
    // e to the B-th power
    //
-   new (Z) FloatCell(exp(APL_Float(value.ival)));
-   return E_NO_ERROR;
+   return FloatCell::zF(Z, exp(APL_Float(value.ival)));
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 ErrorCode
 IntCell::bif_nat_log(Cell * Z) const
 {
@@ -459,24 +455,20 @@ IntCell::bif_nat_log(Cell * Z) const
 
    if (value.ival > 0)
       {
-        new (Z) FloatCell(log(APL_Float(value.ival)));
-      }
-   else
-      {
-        const APL_Complex bb(value.ival, 0);
-        new (Z) ComplexCell(log(bb));
+        return FloatCell::zF(Z, log(APL_Float(value.ival)));
       }
 
-   return E_NO_ERROR;
+const APL_Complex bb(value.ival, 0);
+   return ComplexCell::zC(Z, log(bb));
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 // dyadic built-in functions...
 //
 // where possible a function with non-int A is delegated to the corresponding
 // member function of A. For numeric cells that is the FloatCell or ComplexCell
 // function and otherwise the default function (that returns E_DOMAIN_ERROR.
 //
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 ErrorCode
 IntCell::bif_add(Cell * Z, const Cell * A) const
 {
@@ -488,22 +480,22 @@ IntCell::bif_add(Cell * Z, const Cell * A) const
         const APL_Integer b =    get_int_value();
         const APL_Integer z = a + b;
         if (sum_overflow(z, a, b))
-           return FloatCell::zv(Z, APL_Float(a) + APL_Float(b));
+           return FloatCell::zF(Z, APL_Float(a) + APL_Float(b));
         else
-           return zv(Z,z);
+           return IntCell::zI(Z,z);
       }
 
    // delegate to A
    //
    return A->bif_add(Z, this);
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 ErrorCode
 IntCell::bif_add_inverse(Cell * Z, const Cell * A) const
 {
    return A->bif_subtract(Z, this);
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 ErrorCode
 IntCell::bif_subtract(Cell * Z, const Cell * A) const
 {
@@ -516,9 +508,9 @@ IntCell::bif_subtract(Cell * Z, const Cell * A) const
         const APL_Integer z = a - b;
 
         if (diff_overflow(z, a, b))
-           return FloatCell::zv(Z, APL_Float(a) + APL_Float(b));
+           return FloatCell::zF(Z, APL_Float(a) + APL_Float(b));
         else
-           return zv(Z,z);
+           return IntCell::zI(Z,z);
       }
 
    // delegate to A
@@ -526,7 +518,7 @@ IntCell::bif_subtract(Cell * Z, const Cell * A) const
    this->bif_negative(Z);   // Z = -B
    return A->bif_add(Z, Z);        // Z = A + Z = A + -B
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 ErrorCode
 IntCell::bif_divide(Cell * Z, const Cell * A) const
 {
@@ -540,9 +532,7 @@ const bool A_zero = A->is_near_zero();
         return E_DOMAIN_ERROR;
       }
 
-   if (A_zero)   return IntCell::z0(Z);
-
-   // at this point both A and B are non-zero
+   // at this point B is non-zero
    //
    if (A->is_integer_cell())
       {
@@ -550,21 +540,21 @@ const bool A_zero = A->is_near_zero();
         //
         APL_Integer a = A->get_int_value();
 
-#ifdef RATIONAL_NUMBERS_WANTED
+#ifdef cfg_RATIONAL_NUMBERS_WANTED
         if (b < 0)   // make denominator positive
            {
              a = -a;
              b = -b;
            }
         const APL_Integer gcd = FloatCell::gcd(b, a);
-        if (b == gcd)   return IntCell::zv(Z, a/gcd);
-        return FloatCell::zv(Z, a/gcd, b/gcd);
+        if (b == gcd)   return IntCell::zI(Z, a/gcd);
+        return FloatCell::zR(Z, a/gcd, b/gcd);
 #else
         const APL_Float i_quot = a / b;
         const APL_Float r_quot = a / APL_Float(b);
 
-        if (a != i_quot * b)   return FloatCell::zv(Z, r_quot);
-        else                   return IntCell::zv(Z, i_quot);
+        if (a != i_quot * b)   return FloatCell::zF(Z, r_quot);
+        else                   return FloatCell::zF(Z, i_quot);
 #endif
       }
 
@@ -573,7 +563,7 @@ const bool A_zero = A->is_near_zero();
 const ErrorCode ec = this->bif_reciprocal(Z);   // Z = ÷B
    return ec ? ec : A->bif_multiply(Z, Z);      // Z = A × Z = A × ÷B
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 ErrorCode
 IntCell::bif_multiply(Cell * Z, const Cell * A) const
 {
@@ -586,8 +576,8 @@ IntCell::bif_multiply(Cell * Z, const Cell * A) const
         const APL_Float prod = APL_Float(a) * APL_Float(b);
 
         if (prod > LARGE_INT ||
-            prod < SMALL_INT)   new (Z) FloatCell(prod);
-        else                    new (Z) IntCell(a * b);
+            prod < SMALL_INT)   FloatCell::zF(Z, prod);
+        else                    IntCell::zI(Z, a * b);
         return E_NO_ERROR;
       }
 
@@ -595,13 +585,13 @@ IntCell::bif_multiply(Cell * Z, const Cell * A) const
    //
    return A->bif_multiply(Z, this);
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 ErrorCode
 IntCell::bif_multiply_inverse(Cell * Z, const Cell * A) const
 {
    return A->bif_divide(Z, this);
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 ErrorCode
 IntCell::bif_power(Cell * Z, const Cell * A) const
 {
@@ -615,14 +605,13 @@ const bool invert_Z = b < 0;
 
    // at this point, b ≥ 0
    //
-   if (b <= 1)   // special cases A⋆1, A⋆0, and A⋆¯1
+   if (b <= 1)   // special cases: A⋆1, A⋆0, and A⋆¯1
       {
-        if (b == 0)   return z1(Z);  // A⋆0 is 1
+        if (b == 0)   return IntCell::z1(Z);  // A⋆0 is 1
 
         if (invert_Z)               return A->bif_reciprocal(Z);
         if (A->is_real_cell())      return A->bif_conjugate(Z);
-        new (Z) ComplexCell(A->get_complex_value());
-        return E_NO_ERROR;
+        return ComplexCell::zC(Z, A->get_complex_value());
       }
 
    if (A->is_integer_cell())
@@ -632,13 +621,13 @@ const bool invert_Z = b < 0;
            {
              if (invert_Z)   return E_DOMAIN_ERROR;
 
-             return z0(Z);   // 0^b = 0
+             return IntCell::z0(Z);   // 0^b = 0
            }
-        if (a == 1)    return z1(Z);   // 1^b = 1
+        if (a == 1)    return IntCell::z1(Z);   // 1^b = 1
         if (a == -1)   // -1^b = 1 or -1
            {
-              if (b & 1)   return z_1(Z);   // -1^(2N+2) = -1
-              else         return z1(Z);    // -1^(2N+1) =  1
+              if (b & 1)   return IntCell::z_1(Z);   // -1^(2N+2) = -1
+              else         return IntCell::z1(Z);    // -1^(2N+1) =  1
            }
 
         // try to compute A⋆B as integers
@@ -678,16 +667,10 @@ const bool invert_Z = b < 0;
         if (!overflow)
            {
              if (negate_Z)   zi = - zi;
-             if (invert_Z)
-                {
-                  if (zi == 0)   return E_DOMAIN_ERROR;
-                  new (Z) FloatCell(1.0 / zi);
-                }
-             else
-                {
-                  new (Z) IntCell(zi);
-                }
-             return E_NO_ERROR;
+             if (!invert_Z)   return IntCell::zI(Z, zi);
+
+             if (zi != 0)   return FloatCell::zF(Z, 1.0 / zi);
+             return E_DOMAIN_ERROR;
            }
 
         // A and B are integers, but Z is too big for integers
@@ -698,8 +681,7 @@ const bool invert_Z = b < 0;
 
         if (negate_Z)   z = - z;
         if (invert_Z)   z = 1.0 / z;
-        new (Z) FloatCell(z);
-        return E_NO_ERROR;
+        return FloatCell::zF(Z, z);
       }
 
    if (A->is_float_cell())
@@ -710,13 +692,15 @@ const bool invert_Z = b < 0;
         if (af < 0.0)   af = -af;
 
         APL_Float z = pow(af, bf);
+        if (!isfinite(z))   return E_DOMAIN_ERROR;
+
         if (negate_Z)   z = - z;
         if (invert_Z)
            {
              if (z == 0.0)   return E_DOMAIN_ERROR;
              z = 1.0 / z;
            }
-        return FloatCell::zv(Z, z);
+        return FloatCell::zF(Z, z);
       }
 
    // complex A to the B-th power
@@ -726,14 +710,17 @@ const bool invert_Z = b < 0;
      APL_Complex cb(APL_Float(b), 0.0);
 
      APL_Complex z = complex_power(ca, cb);
+     if (!isfinite(z.real()))   return E_DOMAIN_ERROR;
+     if (!isfinite(z.imag()))   return E_DOMAIN_ERROR;
+
      const APL_Float denom = ComplexCell::mag2(z);
-     if (!invert_Z)   return ComplexCell::zv(Z, z);
+     if (!invert_Z)   return ComplexCell::zC(Z, z);
 
      if (denom == 0.0)   return E_DOMAIN_ERROR;
-     return ComplexCell::zv(Z, z.real() / denom, - z.imag() / denom);
+     return ComplexCell::zC(Z, z.real() / denom, - z.imag() / denom);
    }
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 ErrorCode
 IntCell::bif_residue(Cell * Z, const Cell * A) const
 {
@@ -748,7 +735,7 @@ IntCell::bif_residue(Cell * Z, const Cell * A) const
    // if A is zero , return B
    //
 const APL_Float a = A->get_real_value();
-   if (a == 0.0)   return zv(Z, value.ival);
+   if (a == 0.0)   return IntCell::zI(Z, value.ival);
    if (a > (BIG_INT64_F - 1E10) || a < (1E10 - BIG_INT64_F))
       {
         // for large a compute a⍝b as for floats
@@ -791,59 +778,72 @@ const double qct = Workspace::get_CT();
            {
               if (rest < 0)   rest += A->get_int_value();
            }
-        return zv(Z, rest);
+        return IntCell::zI(Z, rest);
       }
 
 APL_Float z = value.ival - a * qf;
-   if (Cell::is_near_zero(z))   return zv(Z, 0);
+   if (Cell::is_near_zero(z))   return IntCell::z0(Z);
 
    // make ×Z == ×A
    //
    if      (a < 0.0 && z > 0.0)   z = z + a;   // += since a < 0
    else if (a > 0.0 && z < 0.0)   z = z + a;
 
-   return FloatCell::zv(Z, z);
+   return FloatCell::zF(Z, z);
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 ErrorCode
 IntCell::bif_maximum(Cell * Z, const Cell * A) const
 {
    if (A->is_integer_cell())
       {
          const APL_Integer a = A->get_int_value();
-         if (a >= value.ival)   new (Z) IntCell(a);
-         else                   new (Z) IntCell(value.ival);
-         return E_NO_ERROR;
+         if (a >= value.ival)   return IntCell::zI(Z, a);
+         return IntCell::zI(Z, value.ival);
       }
 
    // delegate to A
    //
    return A->bif_maximum(Z, this);
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 ErrorCode
 IntCell::bif_minimum(Cell * Z, const Cell * A) const
 {
    if (A->is_integer_cell())
       {
          const APL_Integer a = A->get_int_value();
-         if (a <= value.ival)   new (Z) IntCell(a);
-         else                   new (Z) IntCell(value.ival);
-         return E_NO_ERROR;
+         if (a <= value.ival)   return IntCell::zI(Z, a);
+         return IntCell::zI(Z, value.ival);
       }
 
    // delegate to A
    //
    return A->bif_minimum(Z, this);
 }
-//=============================================================================
+//----------------------------------------------------------------------------
+/// overloaded Cell::bif_near_int64_t()
+ErrorCode
+IntCell::bif_near_int64_t(Cell * Z) const
+{
+   return IntCell::zI(Z, value.ival);
+}
+
+//----------------------------------------------------------------------------
+/// overloaded Cell::bif_within_quad_CT()
+ErrorCode
+IntCell::bif_within_quad_CT(Cell * Z) const
+{
+   return IntCell::zI(Z, value.ival);
+}
+//============================================================================
 // throw/nothrow boundary. Functions above MUST NOT (directly or indirectly)
 // throw while funcions below MAY throw.
-//=============================================================================
+//============================================================================
 
 #include "Error.hh"
 
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 Comp_result
 IntCell::compare(const Cell & other) const
 {
@@ -858,9 +858,11 @@ IntCell::compare(const Cell & other) const
         return Comp_result(-other.compare(*this));
       }
 
+   if (other.is_character_cell())   return COMP_GT;   // numeric > char
+   if (other.is_pointer_cell())     return COMP_LT;   // numeric < nested
    DOMAIN_ERROR;
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 bool
 IntCell::greater(const Cell & other) const
 {
@@ -893,7 +895,7 @@ const Comp_result comp = compare(other);
    if (comp == COMP_EQ)   return this > &other;
    return (comp == COMP_GT);
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 PrintBuffer
 IntCell::character_representation(const PrintContext & pctx) const
 {
@@ -904,24 +906,26 @@ IntCell::character_representation(const PrintContext & pctx) const
       }
 
 char cc[40];
-int len = snprintf(cc, sizeof(cc), "%lld", long_long(value.ival));
+   SPRINTF(cc, "%lld", long_long(value.ival));
 
 UCS_string ucs;
 
-   loop(l, len)
+   loop(c, sizeof(cc))
        {
-         if (cc[l] == '-')   ucs.append(UNI_OVERBAR);
-         else                ucs.append(Unicode(cc[l]));
+         const char q = cc[c];
+         if (q == 0)   break;
+         if (q == '-')   ucs.append(UNI_OVERBAR);
+         else            ucs.append(Unicode(q));
        }
 
 ColInfo info;
    info.flags |= CT_INT;
    info.real_len = ucs.size();
    info.int_len = ucs.size();
-   
+
    return PrintBuffer(ucs, info);
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 int
 IntCell::get_byte_value() const
 {
@@ -941,7 +945,7 @@ IntCell::get_byte_value() const
 
    return value.ival;
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 int
 IntCell::CDR_size() const
 {
@@ -953,4 +957,4 @@ const APL_Integer val = get_int_value();
    if (val == 0xFFFFFFFF)     return 4;
    return 8;
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
