@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2016  Dr. Jürgen Sauermann
+    Copyright © 2008-2023  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -18,13 +18,16 @@
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
+/** @file
+*/
+
 #include "Error.hh"
 #include "PointerCell.hh"
 #include "PrintOperator.hh"
 #include "Value.hh"
 #include "Workspace.hh"
 
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 PointerCell::PointerCell(Value * sub_val, Value & cell_owner)
 {
    Assert(!sub_val->is_simple_scalar());
@@ -36,7 +39,7 @@ PointerCell::PointerCell(Value * sub_val, Value & cell_owner)
    cell_owner.increment_pointer_cell_count();
    cell_owner.add_subcount(sub_val->nz_element_count());
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 PointerCell::PointerCell(Value * sub_val, Value & cell_owner, uint32_t magic)
 {
    // DO NOT: Assert(!sub_val->is_simple_scalar()); This is a special
@@ -52,30 +55,40 @@ PointerCell::PointerCell(Value * sub_val, Value & cell_owner, uint32_t magic)
    cell_owner.increment_pointer_cell_count();
    cell_owner.add_subcount(sub_val->nz_element_count());
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 void
-PointerCell::init_other(void * other, Value & cell_owner,
+PointerCell::init_other(void * other, Value & other_owner,
                             const char * loc) const
 {
    Assert(other);   // the new PointerCell to be created
 
 Value_P sub;   // instantiate beforehand so that sub is 0 if clone() fails
 
-   sub = get_pointer_value()->clone(loc);
-   Assert(!!sub);
-   new (other) PointerCell(sub.get(), cell_owner);
+   sub = CLONE_P(get_pointer_value(), loc);
+   Assert(+sub);
+   new (other) PointerCell(sub.get(), other_owner);
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 void
 PointerCell::release(const char * loc)
 {
+   // 1. update the PointerCell related counters in the owner of this
+   //    PointerCell (since this Cell is no longer a PointerCell).
+   //
    value.pval.owner->decrement_pointer_cell_count();
    value.pval.owner->add_subcount(-get_pointer_value()->nz_element_count());
 
+   // 2. decrement the owner_count of our sub-value via reset() (since this
+   //    PointerCell releases the ownership of the sub-value.
+   //
    value.pval.valp.reset();
-   new (this) IntCell(0);
+
+   // 3. not needed but to be on the safe side: make this Cell an IntCell
+   //    (in case someone still points to it).
+   //
+   IntCell::z0(this);
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 bool
 PointerCell::equal(const Cell & other, double qct) const
 {
@@ -88,11 +101,11 @@ Value_P B = other.get_pointer_value();
 
 const ShapeItem count = A->nz_element_count();
    loop(c, count)
-       if (!A->get_ravel(c).equal(B->get_ravel(c), qct))   return false;
+       if (!A->get_cravel(c).equal(B->get_cravel(c), qct))   return false;
 
    return true;
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 bool
 PointerCell::greater(const Cell & other) const
 {
@@ -100,20 +113,14 @@ PointerCell::greater(const Cell & other) const
    if (compare(other) == COMP_LT)   return false;
    return this > &other;
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 Comp_result
 PointerCell::compare(const Cell & other) const
 {
-   switch(other.get_cell_type())
-      {
-        case CT_CHAR:
-        case CT_INT:
-        case CT_FLOAT:
-        case CT_COMPLEX: return COMP_GT;
-        case CT_POINTER: break;   // continue below
-        case CT_CELLREF: DOMAIN_ERROR;
-        default:         Assert(0 && "Bad celltype");
-      }
+   if (other.get_cell_type() & CT_SIMPLE)   // nested > numeric > char
+      return COMP_GT;
+
+   if (other.get_cell_type() != CT_POINTER)   DOMAIN_ERROR;
 
    // at this point both cells are pointer cells.
    //
@@ -122,33 +129,33 @@ Value_P v2 = other.get_pointer_value();
 
    // compare ranks
    //
-   if (v1->get_rank() > v2->get_rank())   return COMP_GT;
-   if (v1->get_rank() < v2->get_rank())   return COMP_LT;
+   if (v1->get_rank() != v2->get_rank())   // ranks differ
+      return v1->get_rank() > v2->get_rank() ? COMP_GT : COMP_LT;
 
    // same rank, compare shapes
    //
    loop(r, v1->get_rank())
       {
-        if (v1->get_shape_item(r) > v2->get_shape_item(r))   return COMP_GT;
-        if (v1->get_shape_item(r) < v2->get_shape_item(r))   return COMP_LT;
+        const ShapeItem axis1 = v1->get_shape_item(r);
+        const ShapeItem axis2 = v2->get_shape_item(r);
+        if (axis1 != axis2)   // axis r differs
+           return axis1 > axis2 ? COMP_GT : COMP_LT;
       }
 
    // same rank and shape, compare ravel
    //
-const Cell * C1 = &v1->get_ravel(0);
-const Cell * C2 = &v2->get_ravel(0);
+const Cell * C1 = &v1->get_cfirst();
+const Cell * C2 = &v2->get_cfirst();
    loop(e, v1->nz_element_count())
       {
-        const Comp_result comp = C1++->compare(*C2++);
-        if (comp == COMP_GT)   return  COMP_GT;
-        if (comp == COMP_LT)   return  COMP_LT;
+        if (const Comp_result comp = C1++->compare(*C2++))   return  comp;
       }
 
    // everthing equal
    //
    return COMP_EQ;
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 Value_P
 PointerCell::get_pointer_value() const
 {
@@ -156,28 +163,35 @@ Value * vp = const_cast<Value *>(value.pval.valp.get());
 Value_P ret(vp, LOC);   // Value_P constructor increments owner_count
    return ret;
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
+bool
+PointerCell::is_member_anchor() const
+{
+   return value.pval.valp.value_p &&
+          value.pval.valp.value_p->is_member();
+}
+//----------------------------------------------------------------------------
 CellType
 PointerCell::deep_cell_types() const
 {
    return CellType(CT_POINTER | get_pointer_value()->deep_cell_types());
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 CellType
 PointerCell::deep_cell_subtypes() const
 {
    return CellType(CT_POINTER | get_pointer_value()->deep_cell_subtypes());
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 PrintBuffer
 PointerCell::character_representation(const PrintContext & pctx) const
 {
 Value_P val = get_pointer_value();
-   Assert(!!val);
+   Assert(+val);
 
    if (pctx.get_style() & PST_QUOTE_CHARS)
       {
-        if (val->is_char_vector())   return PrintBuffer(val.getref(), pctx, 0);
+        if (val->is_char_vector())   return PrintBuffer(*val, pctx, 0);
       }
 
    if (pctx.get_style() == PR_APL_FUN)   // APL function display
@@ -190,7 +204,7 @@ Value_P val = get_pointer_value();
              ucs.append(UNI_SINGLE_QUOTE);
              loop(e, ec)
                 {
-                  const Unicode uni = val->get_ravel(e).get_char_value();
+                  const Unicode uni = val->get_cravel(e).get_char_value();
                   ucs.append(uni);
                   if (uni == UNI_SINGLE_QUOTE)   ucs.append(uni);   // ' -> ''
                 }
@@ -198,22 +212,22 @@ Value_P val = get_pointer_value();
            }
         else
            {
-             ucs.append(UNI_ASCII_L_PARENT);
+             ucs.append(UNI_L_PARENT);
              loop(e, ec)
                 {
-                  PrintBuffer pb = val->get_ravel(e).
+                  PrintBuffer pb = val->get_cravel(e).
                         character_representation(pctx);
                   ucs.append(UCS_string(pb, 0, Workspace::get_PW()));
 
-                  if (e < ec - 1)   ucs.append(UNI_ASCII_SPACE);
+                  if (e < ec - 1)   ucs.append(UNI_SPACE);
                 }
-             ucs.append(UNI_ASCII_R_PARENT);
+             ucs.append(UNI_R_PARENT);
            }
 
         ColInfo ci;
         PrintBuffer ret(ucs, ci);
-        ret.get_info().int_len  = ret.get_width(0);
-        ret.get_info().real_len = ret.get_width(0);
+        ret.get_info().int_len  = ret.get_column_count();
+        ret.get_info().real_len = ret.get_column_count();
         return ret;
       }
 
@@ -247,56 +261,32 @@ PrintBuffer ret(*val, pctx, 0);
                    if (sh.get_shape_item(r) == 0)   sh.set_shape_item(r, 1);
                  }
 
-             if (sh.get_volume() == 111)   // one prototype
-                {
-                  ret = PrintBuffer(*proto, pctx, 0);
-                  ret.add_frame(PrintStyle(style), proto->get_shape(),
-                                proto->compute_depth());
-                }
-             else                           // several prototypes
-                {
-                  Value_P proto_reshaped(sh, LOC);
-                  Cell * c = &proto_reshaped->get_ravel(0);
-                  const ShapeItem len = proto_reshaped->nz_element_count();
+             Value_P proto_reshaped(sh, LOC);
+             const ShapeItem len = proto_reshaped->nz_element_count();
 
-                  // store proto in the first ravel item, and copies of proto in
-                  // the subsequent ravel items
-                  //
-                  loop(rv, len)
-                      if (proto->is_simple_scalar())
-                         c++->init(proto->get_ravel(0),
-                                   proto_reshaped.getref(), LOC);
-                      else
-                         new (c++) PointerCell(proto->clone(LOC).get(),
-                                               proto_reshaped.getref());
+             // store proto in the first ravel item, and copies of proto in
+             // the subsequent ravel items
+             //
+             loop(rv, len)   proto_reshaped->next_ravel_Value(proto.get());
 
-                  proto_reshaped->check_value(LOC);
-                  ret = PrintBuffer(*proto_reshaped, pctx, 0);
-                  ret.add_frame(PrintStyle(style), proto_reshaped->get_shape(),
-                                proto_reshaped->compute_depth());
-                }
+             proto_reshaped->check_value(LOC);
+             ret = PrintBuffer(*proto_reshaped, pctx, 0);
+             ret.add_frame(PrintStyle(style), proto_reshaped->get_shape(),
+                           proto_reshaped->compute_depth());
            }
        else
            {
-   if (!(pctx.get_style() & PST_QUOTE_CHARS && val->is_char_array()))
-             ret.add_frame(pctx.get_style(), val->get_shape(),
-                           val->compute_depth());
+             if (!(pctx.get_style() & PST_QUOTE_CHARS && val->is_char_array()))
+                {
+                  ret.add_frame(pctx.get_style(), val->get_shape(),
+                                val->compute_depth());
+                }
            }
       }
 
-   ret.get_info().int_len  = ret.get_width(0);
-   ret.get_info().real_len = ret.get_width(0);
+   ret.get_info().int_len  = ret.get_column_count();
+   ret.get_info().real_len = ret.get_column_count();
    return ret;
 }
-//-----------------------------------------------------------------------------
-void
-PointerCell::to_type()
-{
-Value * sub = get_pointer_value().get();
-   Assert(sub);
-const ShapeItem len = sub->nz_element_count();
-Cell * ravel = &sub->get_ravel(0);
-   loop(l, len)   ravel++->to_type();
-}
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 
