@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2016  Dr. Jürgen Sauermann
+    Copyright © 2008-2023  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -18,6 +18,9 @@
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
+/** @file
+*/
+
 #include <stdio.h>
 #include <string.h>
 
@@ -33,25 +36,25 @@
 /// max sizes for arrays on the stack. Larger values are allocated with new()
 enum
 {
-   PB_MAX_COLS        = 200,
-   PB_MAX_ROWS        = 100,
-   PB_MAX_ITEMS       = PB_MAX_COLS * PB_MAX_ROWS,
-   PB_MAX_BREAKPOINTS = 200,
+   PB_MAX_COLS   = 200,
+   PB_MAX_ROWS   = 100,
+   PB_MAX_ITEMS  = PB_MAX_COLS * PB_MAX_ROWS,
+   PB_MAX_CHUNKS = 200,
 };
 
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 PrintBuffer::PrintBuffer()
    : complete(true)
 {
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 PrintBuffer::PrintBuffer(const UCS_string & ucs, const ColInfo & ci)
    : col_info(ci),
      complete(true)
 {
    buffer.push_back(ucs);
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 PrintBuffer::PrintBuffer(const Value & value, const PrintContext & _pctx,
                          ostream * out)
    : complete(false)
@@ -72,7 +75,7 @@ const ShapeItem ec = value.element_count();
    if (value.is_scalar())
       {
         PERFORMANCE_START(start_1)
-        const Cell & cell = value.get_ravel(0);
+        const Cell & cell = value.get_cfirst();
         PrintContext pctx1(pctx);
         if (cell.need_scaling(pctx))   pctx1.set_scaled();
 
@@ -102,9 +105,9 @@ const ShapeItem ec = value.element_count();
         if (value.is_char_vector())
            {
              UCS_string ucs;
-             ucs.append(UNI_ASCII_DOUBLE_QUOTE);
-             loop(v, ec)   ucs.append(value.get_ravel(v).get_char_value());
-             ucs.append(UNI_ASCII_DOUBLE_QUOTE);
+             ucs.append(UNI_DOUBLE_QUOTE);
+             loop(v, ec)   ucs.append(value.get_cravel(v).get_char_value());
+             ucs.append(UNI_DOUBLE_QUOTE);
              append_ucs(ucs);
              update_info();
              complete = true;
@@ -172,7 +175,7 @@ PrintBuffer * item_matrix = 0;
 
    PERFORMANCE_END(fs_PrintBuffer_B, start_0, ec)
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 void
 PrintBuffer::do_PrintBuffer(const Value & value, const PrintContext & pctx,
                          ostream * out, PrintStyle outer_style,
@@ -198,7 +201,7 @@ const bool nested = !value.is_simple();
            loop(y, rows)
                {
                  if (huge && (ii_count != interrupt_count))   goto interrupted;
-                 if (value.get_ravel(x + y*cols).need_scaling(pctx))
+                 if (value.get_cravel(x + y*cols).need_scaling(pctx))
                     {
                       need_scaling = true;
                       break;
@@ -212,7 +215,7 @@ const bool nested = !value.is_simple();
      //    therefore we have (⍴,value) items. Items are rectangular.
      //
      PERFORMANCE_START(start_2)
-     vector<Rank> max_row_ranks;
+     basic_string<sRank> max_row_ranks;
      max_row_ranks.reserve(rows);
      loop(y, rows)
         {
@@ -226,19 +229,24 @@ const bool nested = !value.is_simple();
                 PrintBuffer & item = item_matrix[y*cols + x];
                 PrintContext pctx1 = pctx;
                 if (scaling[x])   pctx1.set_scaled();
-                const Cell & cell = value.get_ravel(x + y*cols);
+                const Cell & cell = value.get_cravel(x + y*cols);
                 item = cell.character_representation(pctx1);
+                if (!item.get_row_count())
+                   {
+                     UCS_string empty;
+                     item.append_ucs(empty);
+                   }
 
                 if (cell.is_pointer_cell())
                    {
                      Value_P sub_val = cell.get_pointer_value();
-                     const Rank sub_rank = sub_val->get_rank();
+                     const sRank sub_rank = sub_val->get_rank();
                      if (max_row_ranks.back() < sub_rank)
                         max_row_ranks.back() = sub_rank;
                    }
 
-                if (max_row_height < item.get_height())
-                   max_row_height = item.get_height();
+                if (max_row_height < item.get_row_count())
+                   max_row_height = item.get_row_count();
 
                 Assert1(item.is_rectangular());
               }
@@ -302,10 +310,14 @@ const bool nested = !value.is_simple();
             ShapeItem dest_height = 0;
             loop(y, rows)
                {
-                 dest_height += separator_rows(y, value, nested,
-                                               max_row_ranks[y],
-                                               y ? max_row_ranks[y - 1] : 0);
-                 dest_height += item_matrix[y*cols + x].get_height();
+                 const sRank Rk_y_1 = y ? max_row_ranks[y - 1] : 0;
+                 const ShapeItem srows = separator_rows(y, value, nested,
+                                                        max_row_ranks[y],
+                                                        Rk_y_1);
+                 const ShapeItem irows =
+                                 item_matrix[y*cols + x].get_row_count();
+                 Assert(irows);
+                 dest_height += srows + irows;
                }
             dest.buffer.reserve(dest_height);
           }
@@ -316,12 +328,13 @@ const bool nested = !value.is_simple();
 
                 // insert separator row(s)
                 //
-                const ShapeItem srows = separator_rows(y, value, nested,
-                                                       max_row_ranks[y],
-                                                  y ? max_row_ranks[y - 1] : 0);
-                if (srows)   // unless first row
+                if (const ShapeItem srows =
+                                    separator_rows(y, value, nested,
+                                                   max_row_ranks[y],
+                                                  y ? max_row_ranks[y - 1] : 0))
                    {
-                    const UCS_string sepa_row(dest.get_width(0), UNI_iPAD_L0);
+                    const UCS_string sepa_row(dest.get_column_count(),
+                                              UNI_iPAD_L0);
                     loop(r, srows)   dest.append_ucs(sepa_row);
                    }
 
@@ -330,7 +343,7 @@ const bool nested = !value.is_simple();
               }
 
           bool not_char = false;   // determined by get_col_spacing()
-	  const int col_spacing = value.get_col_spacing(not_char, x, framed);
+          const int col_spacing = value.get_col_spacing(not_char, x, framed);
 
           const int32_t max_spacing = (col_spacing > last_col_spacing) 
                                     ?  col_spacing : last_col_spacing;
@@ -342,7 +355,7 @@ const bool nested = !value.is_simple();
              {
                if (last_notchar)
                   {
-                    // the previous column was 'notchar' so we add one pad
+                    // the previous column was 'notchar' so we append one pad
                     // char to it.
                     //
                     pcols[x - 1].pad_r(UNI_iPAD_U2, 1);
@@ -350,19 +363,18 @@ const bool nested = !value.is_simple();
                   }
                else if (not_char)
                   {
-                    // the current column is 'notchar' so we add one pad
+                    // the current column is 'notchar' so we prepend one pad
                     // char to it.
                     //
                     dest.pad_l(UNI_iPAD_U3, 1);
                     ++not_char_spaces;
                   }
 
-               // we want a total spacing of 'max_spacing'
-               // but we deduct the 'not_char_spaces' chars ² and ³
+               // we want a total spacing of 'max_spacing', but we
+               // do not count the 'not_char_spaces' chars ² and ³
                // that were appended above.
                //
-               const int u7_pad_len = max_spacing - not_char_spaces;
-               if (u7_pad_len)
+               if (const int u7_pad_len = max_spacing - not_char_spaces)
                   {
                      pcols[x - 1].pad_r(UNI_iPAD_U7, u7_pad_len);
                   }
@@ -391,9 +403,11 @@ const bool nested = !value.is_simple();
 
      if (!is_rectangular())   // should not happen
         {
-          Q1(get_height())
-          loop(h, get_height())   CERR << "w=" << get_width(h) << "*" << endl;
-          loop(h, get_height())   CERR << "*"  << get_line(h) << "*" << endl;
+          Q1(get_row_count())
+          loop(h, get_row_count())   CERR << "w=" << get_column_count(h)
+                                          << "*" << endl;
+          loop(h, get_row_count())   CERR << "*"  << get_line(h)
+                                          << "*" << endl;
         }
 
      Assert(is_rectangular());
@@ -427,44 +441,54 @@ interrupted:
    clear_interrupt_raised(LOC);
    *out << endl << "INTERRUPT" << endl;
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 void
-PrintBuffer::print_interruptible(ostream & out, Rank rank, int quad_PW)
+PrintBuffer::print_interruptible(ostream & out, sRank rank, int quad_PW)
 {
-   if (get_height() == 0)   return;      // empty PrintBuffer
+   if (get_row_count() == 0)   return;      // empty PrintBuffer
 
-const int total_width = get_width(0);
-const int max_breaks = 2*total_width/quad_PW;
-
-ShapeItem * del = 0;
-ShapeItem __breakpoints[PB_MAX_BREAKPOINTS];
-ShapeItem * breakpoints = __breakpoints;
-   if (max_breaks >= PB_MAX_BREAKPOINTS)
-      breakpoints = del = new ShapeItem[max_breaks];
-ShapeItem bp_len = 0;
-
-   // print rows, breaking at breakpoints
+   // lines may be (very) long (compared to ⎕PW) and if they are then they
+   // need to be broken into 2 or more chunks. A chunk is smaller than ⎕PW;
+   // the first chunk is printed un-indented, while subsequent chunk are
+   // indented by 6 blanks.
    //
-   loop(row, get_height())
-       {
-         if (row)   out << endl;   // end previous row
-         int col = 0;
-         int b = 0;
+const int total_width = get_column_count();
 
-         while (col < total_width)
+basic_string<ShapeItem> chunk_lengths;
+   if (quad_PW)   // APL folding of lines near ⎕PW
+      {
+        const int max_breaks = 2 + total_width/quad_PW;   // a first guess
+
+        chunk_lengths.reserve(max_breaks + 1);
+
+        // initialize chunk_lengths based on the first row of the PrintBuffer.
+        // All subsequent rows are aligned to the first row, therefore the
+        // first row can be taken as a prototype for all rows.
+        //
+        for (int col = 0; col < total_width;)
             {
-              int chunk_len;
-              if (row == 0)   // first row: set up breakpoints
-                 {
-                   chunk_len = get_line(0).compute_chunk_length(quad_PW, col);
-                   breakpoints[bp_len++] = chunk_len;
-                 }
-              else            // subsequent row: re-use breakpoints
-                 {
-                   chunk_len = breakpoints[b++];
-                 }
+              const ShapeItem chunk_len =
+                      get_line(0).compute_chunk_length(quad_PW, col);
+              chunk_lengths.push_back(chunk_len);
+              col += chunk_len;
+            }
+       }
+    else          // no APL wrap around
+       {
+         chunk_lengths.push_back(total_width);
+       }
 
+   // print rows, breaking each row at chunk_lengths
+   //
+   loop(row, get_row_count())
+       {
+         int brk_idx = 0;   // chunk_lengths index
+
+         for (int col = 0; col < total_width;)
+            {
               if (col)   out << endl << "      ";
+
+              const size_t chunk_len = chunk_lengths[brk_idx++];
               UCS_string trow(get_line(row), col, chunk_len);
               trow.remove_trailing_padchars();
 
@@ -481,16 +505,14 @@ ShapeItem bp_len = 0;
                    out << endl << "INTERRUPT" << endl;
                    clear_attention_raised(LOC);
                    clear_interrupt_raised(LOC);
-                   delete del;
                    return;
                  }
             }
-       }
 
-   out << endl;
-   delete del;
+         out << endl;   // end of row
+       }
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 void
 PrintBuffer::pb_empty(const Value & value, PrintContext pctx, 
                              PrintStyle outer_style)
@@ -499,18 +521,19 @@ PrintBuffer::pb_empty(const Value & value, PrintContext pctx,
       {
         if (pctx.get_style() == PR_APL_FUN)
            {
-             if (value.get_ravel(0).is_character_cell())   // ''
+             if (value.get_cfirst().is_character_cell())   // ''
                 {
-                  UCS_string ucs("''");
+                  UCS_string ucs(UTF8_string("''"));
                   ColInfo ci;
                   *this = PrintBuffer(ucs, ci);
                   add_outer_frame(outer_style);
                   return;
                 }
 
-             if (value.get_ravel(0).is_numeric())   // ⍬
+             if (value.get_cfirst().is_numeric())   // ⍬
                 {
-                  UCS_string ucs("⍬");
+                  UTF8_string utf("⍬");
+                  UCS_string ucs(utf);
                   ColInfo ci;
                   *this = PrintBuffer(ucs, ci);
                   add_outer_frame(outer_style);
@@ -524,7 +547,7 @@ PrintBuffer::pb_empty(const Value & value, PrintContext pctx,
         return;   // 1 row
       }
 
-const Shape sh = value.get_shape().without_axis(value.get_rank()-1);
+const Shape sh = value.get_shape().without_last_axis();
    if (sh.get_volume() <= 1)   // empty vector
       {
         add_outer_frame(outer_style);
@@ -540,7 +563,7 @@ ShapeItem lines = sh.get_volume();
    buffer.resize(lines);
    add_outer_frame(outer_style);
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 void
 PrintBuffer::pb_for_function(const Value & value, PrintContext pctx, 
                              PrintStyle outer_style)
@@ -553,7 +576,7 @@ UCS_string ucs;
         ucs.append(UNI_SINGLE_QUOTE);
         loop(e, ec)
            {
-             const Unicode uni = value.get_ravel(e).get_char_value();
+             const Unicode uni = value.get_cravel(e).get_char_value();
              ucs.append(uni);
              if (uni == UNI_SINGLE_QUOTE)   ucs.append(uni);   // ' -> ''
            }
@@ -563,9 +586,9 @@ UCS_string ucs;
       {
         loop(e, ec)
            {
-             PrintBuffer pb = value.get_ravel(e)
+             PrintBuffer pb = value.get_cravel(e)
                                           .character_representation(pctx);
-             if (e)   ucs.append(UNI_ASCII_SPACE);
+             if (e)   ucs.append(UNI_SPACE);
              ucs.append(UCS_string(pb, 0, pctx.get_PW()));
            }
       }
@@ -574,10 +597,10 @@ ColInfo ci;
    *this = PrintBuffer(ucs, ci);
    add_outer_frame(outer_style);
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 ShapeItem
 PrintBuffer::separator_rows(ShapeItem y, const Value & value, bool nested,
-                            Rank rk1, Rank rk2)
+                            sRank rk1, sRank rk2)
 {
    if (y == 0)   return 0;
 
@@ -600,7 +623,7 @@ ShapeItem ret = 0;
 
    return ret;
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 Unicode
 PrintBuffer::get_char(int x, int y) const
 { 
@@ -608,7 +631,7 @@ PrintBuffer::get_char(int x, int y) const
    Assert(x < int(buffer[y].size()));
    return buffer[y][x];
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 void
 PrintBuffer::set_char(int x, int y, Unicode uc)
 {
@@ -616,56 +639,56 @@ PrintBuffer::set_char(int x, int y, Unicode uc)
    Assert(x < int(buffer[y].size()));
    buffer[y][x] = uc;
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 void
 PrintBuffer::pad_l(Unicode pad, ShapeItem count)
 {
    if (count == 1)
       {
-        loop(y, get_height())   buffer[y].prepend(pad);
+        loop(y, get_row_count())   buffer[y].prepend(pad);
       }
    else
       {
         UCS_string pads(count, pad);
-        loop(y, get_height())   buffer[y] = pads + buffer[y];
+        loop(y, get_row_count())   buffer[y] = pads + buffer[y];
       }
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 void
 PrintBuffer::pad_r(Unicode pad, ShapeItem count)
 {
 UCS_string ucs(count, pad);
-   loop(y, get_height())   buffer[y].append(ucs);
+   loop(y, get_row_count())   buffer[y].append(ucs);
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 void
 PrintBuffer::pad_height(Unicode pad, ShapeItem height)
 {
-   if (height > get_height())
+   if (height > get_row_count())
       {
-        UCS_string ucs(get_width(0), pad);
-        while (height > get_height())   buffer.push_back(ucs);
+        UCS_string ucs(get_column_count(), pad);
+        while (height > get_row_count())   buffer.push_back(ucs);
       }
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 void
 PrintBuffer::pad_to_spaces()
 {
-   loop(y, get_height())
-   loop(x, get_width(y))
-      if (is_iPAD_char(get_char(x, y)))   set_char(x, y, UNI_ASCII_SPACE);
+   loop(y, get_row_count())
+   loop(x, get_column_count(y))
+      if (is_iPAD_char(get_char(x, y)))   set_char(x, y, UNI_SPACE);
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 void
 PrintBuffer::pad_height_above(Unicode pad, ShapeItem height)
 {
-   if (height > get_height())
+   if (height > get_row_count())
       {
-        UCS_string ucs(get_width(0), pad);
-        while (height > get_height())   buffer.insert(buffer.begin(), ucs);
+        UCS_string ucs(get_column_count(), pad);
+        while (height > get_row_count())   buffer.insert(0, ucs);
       }
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 void
 PrintBuffer::get_frame_chars(PrintStyle pst,
                              Unicode & HORI, Unicode & VERT,
@@ -675,10 +698,10 @@ PrintBuffer::get_frame_chars(PrintStyle pst,
    switch(pst & PST_CS_MASK)
       {
         case PST_CS_ASCII:
-             HORI = UNI_ASCII_MINUS;
-             VERT = UNI_ASCII_BAR;
-             NW   = UNI_ASCII_FULLSTOP;
-             NE   = UNI_ASCII_FULLSTOP;
+             HORI = UNI_MINUS;
+             VERT = UNI_BAR;
+             NW   = UNI_FULLSTOP;
+             NE   = UNI_FULLSTOP;
              SE   = UNI_SINGLE_QUOTE;
              SW   = UNI_SINGLE_QUOTE;
              break;
@@ -720,7 +743,7 @@ PrintBuffer::get_frame_chars(PrintStyle pst,
              FIXME;
       }
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 void
 PrintBuffer::add_frame(PrintStyle style, const Shape & shape, int depth)
 {
@@ -729,7 +752,7 @@ PrintBuffer::add_frame(PrintStyle style, const Shape & shape, int depth)
 Unicode HORI, VERT, NW, NE, SE, SW;
    get_frame_chars(style, HORI, VERT, NW, NE, SE, SW);
 
-   if (get_height() == 0)   // empty
+   if (get_row_count() == 0)   // empty
       {
         UCS_string upper;
         upper.append(NE);
@@ -745,51 +768,52 @@ Unicode HORI, VERT, NW, NE, SE, SW;
         return;
       }
 
-   // draw | left and right
+   // draw │ left and right
    //
-   loop(y, get_height())
+   loop(y, get_row_count())
       {
         buffer[y].prepend(VERT);
         buffer[y].append(VERT);
 
-        // change internal pad characters to ASCII_SPACE so that they will
+        // change internal pad characters to SPACE so that they will
         // not be removed later and the frame is printed correctly
         //
         buffer[y].map_pad();
 
       }
 
-   // draw - on top and bottom.
+   // draw ─ on top and bottom.
    //
-UCS_string hori(get_width(0), HORI);
+UCS_string hori(get_column_count(), HORI);
 
-   buffer.insert(buffer.begin(), hori);
+   buffer.insert(0, hori);
    buffer.push_back(hori);
 
-   // draw the corners
+   // draw the corners ┌, └, ┐, and ┘
    //
-   set_char(0,                0,                NE);
-   set_char(0,                get_height() - 1, SE);
-   set_char(get_width(0) - 1, 0,                NW);
-   set_char(get_width(0) - 1, get_height() - 1, SW);
+   set_char(0,                      0,                   NE);   // e.g. ╔
+   set_char(0,                      get_row_count() - 1, SE);   // e.g. ╚
+   set_char(get_column_count() - 1, 0,                   NW);   // e.g. ╗
+   set_char(get_column_count() - 1, get_row_count() - 1, SW);   // e.g. ╝
 
    // draw other decorators
    //
-   if (shape.get_rank() > 0)
+   if (shape.get_rank() > 0)               // → on top frame line
       {
         set_char(1, 0, UNI_RIGHT_ARROW);
-        if (style & PST_NARS)
+
+        if (style & PST_NARS)   // numbers indicating axis lengths
            {
              UCS_string ucs;
              ucs.append_number(shape.get_last_shape_item());
-             if (ucs.size() < (get_width(0) - 2))
+             if (ucs.size() < (get_column_count() - 2))
                 {
                   loop(u, ucs.size())   set_char(u + 1, 0, ucs[u]);
                 }
            }
       }
 
-   if (shape.get_rank() > 1)
+   if (shape.get_rank() > 1)               // ↓ on lefyt frame line
       {
         set_char(0, 1, UNI_DOWN_ARROW);
         if (style & PST_NARS)
@@ -800,18 +824,18 @@ UCS_string hori(get_width(0), HORI);
                   if (r)   ucs.append(VERT);
                   ucs.append_number(shape.get_shape_item(r));
                 }
-             if (ucs.size() < (get_height() - 2))
+             if (ucs.size() < (get_row_count() - 2))
                 {
                   loop(u, ucs.size())   set_char(0, u + 1, ucs[u]);
                 }
            }
       }
 
-   if (depth)
+   if (depth)                      // ϵ on bottom frame line
       {
         loop(d, depth - 1)
-            if (d + 1 < get_width(0) - 1)
-               set_char(1 + d, get_height() - 1, UNI_ELEMENT1);
+            if (d + 1 < get_column_count() - 1)
+               set_char(1 + d, get_row_count() - 1, UNI_ELEMENT);
       }
 
    if (style & PST_EMPTY_LAST)   // last (X-) dimension is empty
@@ -826,7 +850,7 @@ UCS_string hori(get_width(0), HORI);
 
    Assert(is_rectangular());
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 void
 PrintBuffer::add_outer_frame(PrintStyle style)
 {
@@ -836,7 +860,7 @@ PrintBuffer::add_outer_frame(PrintStyle style)
 Unicode HORI, VERT, NW, NE, SE, SW;
    get_frame_chars(style, HORI, VERT, NW, NE, SE, SW);
 
-   if (get_height() == 0)   // empty
+   if (get_row_count() == 0)   // empty
       {
         UCS_string upper;
         upper.append(NE);
@@ -854,12 +878,12 @@ Unicode HORI, VERT, NW, NE, SE, SW;
 
    // draw a bar left and right
    //
-   loop(y, get_height())
+   loop(y, get_row_count())
       {
         buffer[y].prepend(VERT);
         buffer[y].append(VERT);
 
-        // change internal pad characters to ASCII_SPACE so that they will
+        // change internal pad characters to SPACE so that they will
         // not be removed later and the frame is printed correctly
         //
         buffer[y].map_pad();
@@ -867,27 +891,27 @@ Unicode HORI, VERT, NW, NE, SE, SW;
 
    // draw a bar on top and bottom.
    //
-UCS_string hori(get_width(0), HORI);
+UCS_string hori(get_column_count(), HORI);
 
-   buffer.insert(buffer.begin(), hori);
+   buffer.insert(0, hori);
    buffer.push_back(hori);
 
    // draw the corners
    //
-   set_char(0,                0,                NE);
-   set_char(0,                get_height() - 1, SE);
-   set_char(get_width(0) - 1, 0,                NW);
-   set_char(get_width(0) - 1, get_height() - 1, SW);
+   set_char(0,                       0,                   NE);
+   set_char(0,                       get_row_count() - 1, SE);
+   set_char(get_column_count() - 1, 0,                   NW);
+   set_char(get_column_count() - 1, get_row_count() - 1, SW);
 
    Assert(is_rectangular());
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 ostream &
 PrintBuffer::debug(ostream & out, const char * title) const
 {
    if (title)   out << title << endl;
 
-   if (get_height() == 0)
+   if (get_row_count() == 0)
       {
         out << UNI_LINE_DOWN_RIGHT << UNI_LINE_DOWN_LEFT << endl
             << UNI_LINE_UP_RIGHT   << UNI_LINE_UP_LEFT
@@ -899,14 +923,14 @@ PrintBuffer::debug(ostream & out, const char * title) const
       }
 
    out << UNI_LINE_DOWN_RIGHT;
-   loop(w, get_width(0))   out << UNI_LINE_HORI;
+   loop(w, get_column_count())   out << UNI_LINE_HORI;
    out << UNI_LINE_DOWN_LEFT << endl;
 
-   loop(y, get_height())
+   loop(y, get_row_count())
    out << UNI_LINE_VERT << buffer[y] << UNI_LINE_VERT << endl;
 
    out << UNI_LINE_UP_RIGHT;
-   loop(w, get_width(0))   out << UNI_LINE_HORI;
+   loop(w, get_column_count())   out << UNI_LINE_HORI;
    out << UNI_LINE_UP_LEFT
        << " flg=" << HEX(col_info.flags)
        << " il="  << col_info.int_len
@@ -917,41 +941,45 @@ PrintBuffer::debug(ostream & out, const char * title) const
 
    return out;
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 void
 PrintBuffer::append_col(const PrintBuffer & pb1)
 {
-   Assert(get_height() == pb1.get_height());
+   Assert(get_row_count() == pb1.get_row_count());
 
-   loop(h, get_height())   buffer[h].append(pb1.buffer[h]);
+   loop(h, get_row_count())   buffer[h].append(pb1.buffer[h]);
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 void
 PrintBuffer::append_ucs(const UCS_string & ucs)
 {
-   if (buffer.size() == 0)   // empty
+   if (buffer.size() == 0)   // empty buffer (no lines yet) : add ucs
       {
         buffer.push_back(ucs);
+        return;
       }
-   else if (ucs.size() < get_width(0))
+
+const int size = ucs.size();
+   if (size < get_column_count())  // new line is shorter: pad it)
       {
         UCS_string ucs1(ucs);
-        UCS_string pad(get_width(0) - ucs.size(), UNI_iPAD_L1);
+        UCS_string pad(get_column_count() - size, UNI_iPAD_L1);
         ucs1.append(pad);
         buffer.push_back(ucs1);
+        return;
       }
-   else if (ucs.size() > get_width(0))
-      {
-        UCS_string pad(ucs.size() - get_width(0), UNI_iPAD_L2);
-        loop(h, get_height())   buffer[h].append(pad);
-        buffer.push_back(ucs);
-      }
-   else
+
+   if (size > get_column_count())   // new line is longer: pad PrintBufer
       {
+        UCS_string pad(ucs.size() - get_column_count(), UNI_iPAD_L2);
+        loop(h, get_row_count())   buffer[h].append(pad);
         buffer.push_back(ucs);
+        return;
       }
+
+   buffer.push_back(ucs);
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 void
 PrintBuffer::append_aligned(const UCS_string & ucs, Unicode align)
 {
@@ -983,7 +1011,7 @@ int this_pos = -1;
       }
 
 const int ucs_w = ucs.size();
-const int this_w = get_width(0);
+const int this_w = get_column_count();
 
 int ucs_l = 0;    // padding left of ucs
 int ucs_r = 0;    // padding right of ucs
@@ -1043,44 +1071,44 @@ int this_r = 0;   // padding right of this
 
 UCS_string ucs1;
 
-   if (ucs_l > 0)   ucs1.append(UCS_string(ucs_l, UNI_ASCII_SPACE));
+   if (ucs_l > 0)   ucs1.append(UCS_string(ucs_l, UNI_SPACE));
    ucs1.append(ucs);
-   if (ucs_r > 0)   ucs1.append(UCS_string(ucs_r, UNI_ASCII_SPACE));
+   if (ucs_r > 0)   ucs1.append(UCS_string(ucs_r, UNI_SPACE));
 
-   if (this_l > 0)   pad_l(UNI_ASCII_SPACE, this_l);
-   if (this_r > 0)   pad_r(UNI_ASCII_SPACE, this_r);
+   if (this_l > 0)   pad_l(UNI_SPACE, this_l);
+   if (this_r > 0)   pad_r(UNI_SPACE, this_r);
 
    buffer.push_back(ucs1);
 
    Assert(is_rectangular());
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 void
 PrintBuffer::add_column(Unicode pad, int32_t pad_count, const PrintBuffer & pb)
 {
-   if (get_height() != pb.get_height())
+   if (get_row_count() != pb.get_row_count())
       {
          debug(CERR, "this");
          pb.debug(CERR, "pb");
       }
 
-   Assert(get_height() == pb.get_height());
+   Assert(get_row_count() == pb.get_row_count());
 
    if (pad_count)
       {
         UCS_string ucs(pad_count, pad);
-        loop(y, get_height())   buffer[y].append(ucs);
+        loop(y, get_row_count())   buffer[y].append(ucs);
       }
 
-   loop(y, get_height())   buffer[y].append(pb.buffer[y]);
+   loop(y, get_row_count())   buffer[y].append(pb.buffer[y]);
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 void PrintBuffer::add_row(const PrintBuffer & pb)
 {
-   buffer.reserve(buffer.size() + pb.get_height());
-   loop(h, pb.get_height())   buffer.push_back(pb.buffer[h]);
+   buffer.reserve(buffer.size() + pb.get_row_count());
+   loop(h, pb.get_row_count())   buffer.push_back(pb.buffer[h]);
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 void
 PrintBuffer::align(ColInfo & cols)
 {
@@ -1101,7 +1129,7 @@ PrintBuffer::align(ColInfo & cols)
 
    Assert(is_rectangular());
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 void
 PrintBuffer::align_dot(ColInfo & COL_INFO)
 {
@@ -1176,8 +1204,8 @@ PrintBuffer::align_dot(ColInfo & COL_INFO)
                   else                        // no expo yet: create one
                      {
                        Assert1(diff >= 2);
-                       pad_r(UNI_ASCII_E, 1);
-                       pad_r(UNI_ASCII_0, 1);
+                       pad_r(UNI_E, 1);
+                       pad_r(UNI_0, 1);
                        pad_r(UNI_iPAD_L4, diff - 2);
                      }
                   col_info.real_len = COL_INFO.real_len;
@@ -1202,7 +1230,7 @@ PrintBuffer::align_dot(ColInfo & COL_INFO)
 
    Log(LOG_printbuf_align)   debug(CERR, "after align_dot()");
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 void
 PrintBuffer::align_j(ColInfo & COL_INFO)
 {
@@ -1258,7 +1286,7 @@ PrintBuffer::align_j(ColInfo & COL_INFO)
 
    Log(LOG_printbuf_align)   debug(CERR, "after align_j()");
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 void
 PrintBuffer::pad_fraction(int wanted_fract_len, bool want_expo)
 {
@@ -1281,12 +1309,12 @@ UCS_string new_buf(buffer[0], 0, col_info.int_len);
          }
       else if (col_info.fract_len == 0)   // no fractional part (yet), e.g. 1E2
          {
-           new_buf.append(UNI_ASCII_FULLSTOP);
-           loop(d, diff - 1)   new_buf.append(UNI_ASCII_0);
+           new_buf.append(UNI_FULLSTOP);
+           loop(d, diff - 1)   new_buf.append(UNI_0);
          }
       else
          {
-           loop(d, diff)   new_buf.append(UNI_ASCII_0);
+           loop(d, diff)   new_buf.append(UNI_0);
          }
 
    // copy exponent part
@@ -1302,9 +1330,9 @@ UCS_string new_buf(buffer[0], 0, col_info.int_len);
 
    // if buffer is multi-len then pad remaining line to new length
    //
-   for (ShapeItem h = 1; h < get_height(); ++h)
+   for (ShapeItem h = 1; h < get_row_count(); ++h)
       {
-        const int diff = new_buf.size() - get_width(h);
+        const int diff = new_buf.size() - get_column_count(h);
         if (diff > 0)
            {
              const UCS_string ucs(diff, UNI_iPAD_L4);
@@ -1312,7 +1340,7 @@ UCS_string new_buf(buffer[0], 0, col_info.int_len);
            }
       }
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 void
 PrintBuffer::align_left(ColInfo & COL_INFO)
 {
@@ -1340,22 +1368,22 @@ const size_t diff = COL_INFO.int_len - col_info.int_len;
 
    Log(LOG_printbuf_align)   debug(CERR, "after align_left()");
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 bool
 PrintBuffer::is_rectangular() const
 {
-   if (get_height())
+   if (get_row_count())
       {
-        const ShapeItem w = get_width(0);
-        loop(h, get_height())
+        const ShapeItem w = get_column_count();
+        loop(h, get_row_count())
            {
-             if (get_width(h) != w)    return false;
+             if (get_column_count(h) != w)    return false;
            }
       }
 
    return true;
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 void
 ColInfo::consider(const ColInfo & item)
 {
@@ -1393,10 +1421,10 @@ ColInfo::consider(const ColInfo & item)
 
    if (imag_len  < item.imag_len)    imag_len  = item.imag_len;
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 ostream &
 operator << (ostream & out, const PrintBuffer & pb)
 {
    out << endl;   return pb.debug(out);
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
