@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2016  Dr. Jürgen Sauermann
+    Copyright © 2008-2023  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -18,6 +18,9 @@
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
+/** @file
+*/
+
 #include <stdlib.h>
 #include <dlfcn.h>
 
@@ -30,19 +33,18 @@
 #include "UserPreferences.hh"
 #include "Workspace.hh"
 
-Quad_FX  Quad_FX::_fun;
-Quad_FX * Quad_FX::fun = &Quad_FX::_fun;
+Quad_FX  Quad_FX::fun;
 
-//=============================================================================
+//============================================================================
 Token
-Quad_FX::eval_B(Value_P B)
+Quad_FX::do_eval_B(const Value * B)
 {
 static const int eprops[] = { 0, 0, 0, 0 };
    return do_quad_FX(eprops, B, UTF8_string("⎕FX"), false);
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 Token
-Quad_FX::eval_AB(Value_P A, Value_P B)
+Quad_FX::do_eval_AB(const Value * A, const Value * B)
 {
    CHECK_SECURITY(disable_native_functions);
 
@@ -60,7 +62,8 @@ Quad_FX::eval_AB(Value_P A, Value_P B)
    // 2d.  N1 N2 N3 N4 "creator" ⎕FX "APL-text"   N1 N2 N3 N4      "creator"
    //
 
-   if (A->is_char_string())   return do_native_FX(A, -1, B);
+   if (A->is_char_string())
+      return Token(TOK_APL_VALUE1, do_native_FX(A, -1, B));
 
 int eprops[4];
 UTF8_string creator("⎕FX");
@@ -69,13 +72,13 @@ UTF8_string creator("⎕FX");
       {
         case 2:   // format 2b.
              {
-               const Value & C = *A->get_ravel(1).get_pointer_value().get();
+               const Value & C = *A->get_cravel(1).get_pointer_value().get();
                UCS_string creator_ucs(C);
                creator = UTF8_string(creator_ucs);
              }
              /* no break */
         case 1:   // format 2a.
-             eprops[0] = A->get_ravel(0).get_int_value();
+             eprops[0] = A->get_cfirst().get_int_value();
              if (eprops[0] < 0)   DOMAIN_ERROR;
              if (eprops[0] > 1)   DOMAIN_ERROR;
              eprops[3] = eprops[2] = eprops[1] = eprops[0];
@@ -83,7 +86,7 @@ UTF8_string creator("⎕FX");
 
         case 5:   // format 2d.
              {
-               const Value & C = *A->get_ravel(4).get_pointer_value().get();
+               const Value & C = *A->get_cravel(4).get_pointer_value().get();
                UCS_string creator_ucs(C);
                creator = UTF8_string(creator_ucs);
              }
@@ -91,7 +94,7 @@ UTF8_string creator("⎕FX");
         case 4:   // format 2c.
              loop(e, 4)
                 {
-                  eprops[e] = A->get_ravel(e).get_int_value();
+                  eprops[e] = A->get_cravel(e).get_int_value();
                   if (eprops[e] < 0)   DOMAIN_ERROR;
                   if (eprops[e] > 1)   DOMAIN_ERROR;
                 }
@@ -102,9 +105,9 @@ UTF8_string creator("⎕FX");
 
    return do_quad_FX(eprops, B, creator, false);
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 Token
-Quad_FX::eval_AXB(Value_P A, Value_P X, Value_P B)
+Quad_FX::eval_AXB(Value_P A, Value_P X, Value_P B) const
 {
    CHECK_SECURITY(disable_native_functions);
 
@@ -112,12 +115,12 @@ Quad_FX::eval_AXB(Value_P A, Value_P X, Value_P B)
    if (A->get_rank() > 1)                RANK_ERROR;
    if (!A->is_char_string())             DOMAIN_ERROR;
 
-const Axis axis = Value::get_single_axis(X.get(), 10);
-   return do_native_FX(A, axis, B);
+const sAxis axis = Value::get_single_axis(X.get(), 10);
+   return Token(TOK_APL_VALUE1, do_native_FX(A.get(), axis, B.get()));
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 Token
-Quad_FX::do_quad_FX(const int * exec_props, Value_P B,
+Quad_FX::do_quad_FX(const int * exec_props, const Value * B,
                     const UTF8_string & creator, bool tolerant)
 {
    if (B->get_rank() > 2)   RANK_ERROR;
@@ -127,7 +130,7 @@ UCS_string text;
 
    // ⎕FX accepts two kinds of B argments:
    //
-   // 1. A vector whose elements are the lines of the function, or
+   // 1. A vector whose elements are the (nested) lines of the function, or
    // 2. A character matrix whose rows are the lines of the function.
    //
    // we convert each format into text, which is a UCS string with
@@ -135,66 +138,97 @@ UCS_string text;
    //
    if (B->compute_depth() >= 2)   // case 1: vector of simple character vectors
       {
-        const ShapeItem ec = B->element_count();
-        loop(e, ec)
+        const ShapeItem rows = B->element_count();
+        loop(row, rows)
            {
-             const Cell & cell = B->get_ravel(e);
+             const Cell & cell = B->get_cravel(row);
              if (cell.is_character_cell())   /// a line with a single char.
                 {
-                  text.append(cell.get_char_value());
-                  text.append(UNI_ASCII_LF);
+                  const Unicode uni = cell.get_char_value();
+                  if (uni > UNI_SPACE)   text.append(uni);
+                  text.append(UNI_LF);
                   continue;
                 }
 
-             if (!cell.is_pointer_cell())   DOMAIN_ERROR;
+             // row has more than 1 chatacter, so it must be nested
+             if (!cell.is_pointer_cell())
+                {
+                  MORE_ERROR() << "⎕FX: Function line " << row
+                               << " is not a string";
+                  DOMAIN_ERROR;
+                }
+
              Value_P line = cell.get_pointer_value();
-             Assert(!!line);
+             Assert(+line);
 
              Log(LOG_quad_FX)
                 {
-                  CERR << "[" << setw(2) << e << "] " << *line << endl;
+                  CERR << "[" << setw(2) << row << "] " << *line << endl;
                 }
 
              if (line->is_char_vector())
                 {
-                  const uint64_t line_len = line->element_count();
+                  const ShapeItem line_len = line->element_count();
+                  bool skipping = false;
                   loop(l, line_len)
                      {
-                       const Cell & c = line->get_ravel(l);
-                       if (!c.is_character_cell())   DOMAIN_ERROR;
-                       text.append(c.get_char_value());
+                       const Cell & c = line->get_cravel(l);
+                       if (!c.is_character_cell())
+                          {
+                            MORE_ERROR() << "non-char in line at " LOC;
+                            DOMAIN_ERROR;
+                          }
+
+                       const Unicode uni = c.get_char_value();
+                       if (l == 0 || skipping)
+                          skipping = (uni <= UNI_SPACE);
+                       if (!skipping)   text.append(uni);
                      }
                 }
              else if (line->is_scalar())
                 {
-                  const Cell & c1 = line->get_ravel(0);
-                  if (!c1.is_character_cell())   DOMAIN_ERROR;
-                  text.append(c1.get_char_value());
+                  const Cell & c1 = line->get_cfirst();
+                  if (!c1.is_character_cell())
+                     {
+                       MORE_ERROR() << "non-char in line at " LOC;
+                       DOMAIN_ERROR;
+                     }
+                  const Unicode uni = c1.get_char_value();
+                  if (uni > UNI_SPACE)   text.append(uni);
                 }
              else
                 {
+                  line->print_boxed(CERR, 0);
+                  MORE_ERROR() << "bad line at " LOC;
                   DOMAIN_ERROR;
                 }
 
-             text.append(UNI_ASCII_LF);
+             text.append(UNI_LF);
            }
       }
    else                      // case 2: simple character matrix
       {
         const ShapeItem rows = B->get_rows();
         const ShapeItem cols = B->get_cols();
-        const Cell * cB = &B->get_ravel(0);
+        const Cell * cB = &B->get_cfirst();
 
         loop(row, rows)
            {
-             loop(col, cols)   text.append(cB++->get_char_value());
-             text.append(UNI_ASCII_LF);
+             bool skipping = false;
+             loop(col, cols)
+                 {
+                   const Unicode uni = cB++->get_char_value();
+                   if (col == 0 || skipping)
+                      skipping = (uni <= UNI_SPACE);
+                   if (!skipping)   text.append(uni);
+                 }
+             text.append(UNI_LF);
            }
       }
 
    return do_quad_FX(exec_props, text, creator, tolerant);
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 Token
 Quad_FX::do_quad_FX(const int * exec_props, const UCS_string & text,
                     const UTF8_string & creator, bool tolerant)
@@ -205,7 +239,7 @@ UserFunction * fun = UserFunction::fix(text, error_line, false, LOC, creator,
    if (fun == 0)
       {
         Value_P Z(LOC);
-        new (Z->next_ravel())   IntCell(error_line + Workspace::get_IO());
+        Z->next_ravel_Int(error_line + Workspace::get_IO());
         Z->check_value(LOC);
         return Token(TOK_APL_VALUE1, Z);
       }
@@ -215,26 +249,24 @@ UserFunction * fun = UserFunction::fix(text, error_line, false, LOC, creator,
 const UCS_string fun_name = fun->get_name();
 Value_P Z(fun_name, LOC);
 
-        Z->check_value(LOC);
-        return Token(TOK_APL_VALUE1, Z);
+   Z->check_value(LOC);
+   return Token(TOK_APL_VALUE1, Z);
 }
-//-----------------------------------------------------------------------------
-Token
-Quad_FX::do_native_FX(Value_P A, Axis axis, Value_P B)
+//----------------------------------------------------------------------------
+Value_P
+Quad_FX::do_native_FX(const Value * A, sAxis axis, const Value * B)
 {
-   if (uprefs.safe_mode)   DOMAIN_ERROR;
+   if (UserPreferences::uprefs.safe_mode)   DOMAIN_ERROR;
 
-UCS_string so_name       = A->get_UCS_ravel();
-UCS_string function_name = B->get_UCS_ravel();
+const UCS_string so_name       = A->get_UCS_ravel();
+const UCS_string function_name = B->get_UCS_ravel();
 
    if (so_name.size() == 0)         LENGTH_ERROR;
    if (function_name.size() == 0)   LENGTH_ERROR;
 
 NativeFunction * fun = NativeFunction::fix(so_name, function_name);
-   if (fun == 0)  return Token(TOK_APL_VALUE1, IntScalar(0, LOC));
+   if (fun == 0)  return IntScalar(0, LOC);
 
-Value_P Z = B;
-   Z->check_value(LOC);
-   return Token(TOK_APL_VALUE1, Z);
+   return CLONE(B, LOC);
 }
-//=============================================================================
+//============================================================================
