@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
+    Copyright © 2008-2023  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -18,6 +18,9 @@
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
+/** @file
+*/
+
 #include <stdlib.h>
 
 #include "Quad_RL.hh"
@@ -26,49 +29,49 @@
 
 uint64_t Quad_RL::state = 0;
 
-//=============================================================================
+//============================================================================
 Quad_RL::Quad_RL()
    : SystemVariable(ID_Quad_RL)
 {
-Value_P value(LOC);
+Value_P Z(LOC);
 
 const unsigned int seed = reset_seed();
 
-   new (value->next_ravel()) IntCell(seed);
-   value->check_value(LOC);
+   Z->next_ravel_Int(seed);
+   Z->check_value(LOC);
 
-   Symbol::assign(value, false, LOC);
+   Symbol::assign(Z, false, LOC);
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 void
-Quad_RL::assign(Value_P value, bool clone, const char * loc)
+Quad_RL::assign(Value_P B, bool /* clone */, const char * loc)
 {
-   if (!value->is_scalar_or_len1_vector())
+   if (!B->is_scalar_or_len1_vector())
       {
-        if (value->get_rank() > 1)   RANK_ERROR;
-        else                         LENGTH_ERROR;
+        if (B->get_rank() > 1)   RANK_ERROR;
+        else                     LENGTH_ERROR;
       }
 
-const Cell & cell = value->get_ravel(0);
+const Cell & cell = B->get_cscalar();
 const APL_Integer val = cell.get_near_int();
 
    state = val;
-   Symbol::assign(IntScalar(val, LOC), false, LOC);
+   Symbol::assign(IntScalar(state, loc), false, LOC);
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 uint64_t
 Quad_RL::get_random()
 {
+   Assert(value_stack.size());   // by Quad_RL::assign()
+   if (value_stack.back().get_NC() != NC_VARIABLE)   VALUE_ERROR;  // localized
+
    state *= Knuth_a;
    state += Knuth_c;
 
-   Assert(value_stack.size());
-   if (value_stack.back().name_class != NC_VARIABLE)   VALUE_ERROR;
-
-   new (&value_stack.back().apl_val->get_ravel(0))   IntCell(state);
+   value_stack.back().get_val_wptr()->set_ravel_Int(0, state);
    return state;
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 void
 Quad_RL::push()
 {
@@ -76,11 +79,11 @@ Quad_RL::push()
    //
    Symbol::push_value(IntScalar(state, LOC));
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 void
 Quad_RL::pop()
 {
    Symbol::pop();
-   state = value_stack.back().apl_val->get_ravel(0).get_near_int();
+   state = value_stack.back().get_val_cptr()->get_cfirst().get_near_int();
 }
-//=============================================================================
+//============================================================================
