@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2016  Dr. Jürgen Sauermann
+    Copyright © 2008-2023  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -18,6 +18,9 @@
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
+/** @file
+*/
+
 #include <string.h>
 
 #include "IndexExpr.hh"
@@ -27,48 +30,51 @@
 #include "Token.hh"
 #include "Workspace.hh"
 
-//-----------------------------------------------------------------------------
-Token::Token(const Token & other)
-   : tag(TOK_VOID)
+//----------------------------------------------------------------------------
+Token::Token(TokenTag t, IndexExpr & idx)
 {
-   copy_1(other, "Token::Token(other)");
-}
-//-----------------------------------------------------------------------------
-Token::Token(TokenTag tg, IndexExpr & idx)
-{
-   if (tg == TOK_PINDEX)
+   /* construct a token for an index expressions. There are 3 cases which
+      differ by tag and by value type:
+
+      1. t indicates a partial index. Partial indices are used after the
+         trailing ] but before the leading [ was parsed (remember that we
+         parse right to left). Tag is TOK_PINDEX / TV_INDEX
+
+      2. t indicates a complete token (after parsing [). There are 2 sub-cases:
+
+      2a. idx has rank 1: [ axis ]. IndexExpr is converted to a (single) value,
+          possibly 0 to indicate an elided index []. Tag is TOK_AXIS / TV_VAL
+
+      2b. idx is [i1;i2...]: Tag is TOK_INDEX / TV_INDEX.
+
+      In case 1. and 2b. the tag == t; only in case 2a. is the caller's
+      t == TOK_INDEX replaced with tok == TOK_AXIS.
+    */
+
+   if (t == TOK_PINDEX)   // case 1. (partial index   ...] )
       {
         // this token is a partial index in the prefix parser.
         // use it as is.
         //
         tag = TOK_PINDEX;
-        value.index_val = &idx;
+        value.index_val = &idx;        // type is TV_INDEX
       }
-   else
+   else                   // case 2. (complete index [ ... ] )
       {
-        // this token is a complete index
-        //
-        Assert(tg == TOK_INDEX);
-        if (idx.value_count() < 2)   // [idx] or []
+        Assert(t == TOK_INDEX);
+        if (idx.get_rank() < 2)   // case 2a. [idx] or []
            {
-             tag = TOK_AXES;
-             if (idx.value_count() == 0)   // []
-                {
-                  new (&value.apl_val) Value_P;
-                }
-             else                          // [x]
-                {
-                  new (&value.apl_val)   Value_P(idx.extract_value(0));
-                }
+             tag = TOK_AXIS;
+             value.apl_val = idx.extract_axis();   // type is TV_VAL
            }
-        else                         // [idx; ...]
+        else                           // [idx; ...]
            {
              tag = TOK_INDEX;
-             value.index_val = &idx;
+             value.index_val = &idx;   // type is TV_INDEX
            }
       }
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 ostream &
 operator << (ostream & out, TokenTag tag)
 {
@@ -77,7 +83,7 @@ operator << (ostream & out, TokenTag tag)
 
    return out;
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 ostream &
 operator << (ostream & out, TokenClass tc)
 {
@@ -118,7 +124,7 @@ operator << (ostream & out, TokenClass tc)
 
    return out;
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 ostream &
 operator << (ostream & out, const Token & token)
 {
@@ -137,17 +143,17 @@ operator << (ostream & out, const Token & token)
         else if (token.get_tag() == TOK_APL_VALUE3)   out << "VALUE3";
         else                                          out << "VALUE???";
         Value_P value = token.get_apl_val();
-        Assert(!!value);
-        const Depth depth = value->compute_depth();
+        Assert(+value);
+        const APL_types::Depth depth = value->compute_depth();
         out << "«";
-        for (Depth d = 0; d < depth; ++d)   out << "≡";
+        for (APL_types::Depth d = 0; d < depth; ++d)   out << "≡";
 
         if (value->get_rank())   out << value->get_shape();
 
         const PrintContext pctx(PR_APL);
         PrintBuffer pb(*value, pctx, 0);
-        bool more = pb.get_height() > 1;
-        if (pb.get_height() > 0)
+        bool more = pb.get_row_count() > 1;
+        if (pb.get_row_count() > 0)
            {
              UCS_string ucs = pb.get_line(0).no_pad();
              if (ucs.size() > 20)
@@ -180,16 +186,18 @@ operator << (ostream & out, const Token & token)
 
    switch(token.get_Class())
       {
-        case TC_END:      return out << "END";
+        case TC_END:
+             return out << "END";
+
         case TC_RETURN:
              if (token.get_tag() == TOK_RETURN_EXEC)
-                          return out << "RETURN ⍎";
+                return out << "RETURN ⍎";
              if (token.get_tag() == TOK_RETURN_STATS)
-                          return out << "RETURN ◊";
+                return out << "RETURN ◊";
              if (token.get_tag() == TOK_RETURN_VOID)
-                          return out << "RETURN ∇FUN";
+                return out << "RETURN ∇FUN";
              if (token.get_tag() == TOK_RETURN_SYMBOL)
-                          return out << "RETURN Z←FUN";
+                return out << "RETURN Z←FUN";
              return out << "RETURN ???";
 
         case TC_VALUE:    return token.print_value(out);
@@ -198,10 +206,14 @@ operator << (ostream & out, const Token & token)
         case TC_PINDEX:
              if (token.get_tag() == TOK_INDEX)
                 return out << token.get_index_val();
-             else if (token.get_tag() == TOK_AXES)
-                return out << *token.get_axes();
-             else
-                FIXME;
+             if (token.get_tag() == TOK_AXIS)
+                {
+                  out << "[";
+                  if (const Value * val = token.get_axes().get())
+                     out << *val;
+                  return out << "]";
+                }
+             FIXME;
 
         case TC_SYMBOL:
              if (token.get_tag() == TOK_LSYMB)
@@ -233,7 +245,7 @@ operator << (ostream & out, const Token & token)
 
    return out <<  "{-unknown Token " << token.get_tag() << "-}";
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 void
 Token::ChangeTag(TokenTag new_tag)
 {
@@ -241,21 +253,27 @@ Token::ChangeTag(TokenTag new_tag)
    // tag is ia const TokenTag, so we cheat a little here.
    const_cast<TokenTag &>(tag) = new_tag;
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 int
 Token::value_use_count() const
 {
-   if (!is_apl_val())    return 0;
-   if (!value.apl_val)   return -98;
-   return value.apl_val->get_owner_count();
+   if (!is_apl_val())    return 0;            // token is not a TV_VALUE token
+   if (!value.apl_val)   return -98;          // it is, but its Value * is -
+   return value.apl_val->get_owner_count();   // non-zero Value *
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
+void
+Token::release_apl_val(const char * loc)
+{
+   if (is_apl_val())   value.apl_val.reset();
+}
+//----------------------------------------------------------------------------
 void
 Token::extract_apl_val(const char * loc)
 {
    if (is_apl_val())   value.apl_val.reset();
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 Value *
 Token::extract_and_keep(const char * loc)
 {
@@ -265,7 +283,7 @@ Value * ret = value.apl_val.get();
    value.apl_val.clear_pointer(loc);
    return ret;
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 ostream &
 Token::print_function(ostream & out) const
 {
@@ -298,26 +316,26 @@ Token::print_function(ostream & out) const
 
         case TOK_F1_EXECUTE:    return out << UNI_EXECUTE;
 
-        case TOK_F2_LESS:       return out << UNI_ASCII_LESS;
+        case TOK_F2_LESS:       return out << UNI_LESS;
         case TOK_F2_FIND:       return out << UNI_EPSILON_UBAR;
-        case TOK_F2_EQUAL:      return out << UNI_ASCII_EQUAL;
-        case TOK_F2_GREATER:    return out << UNI_ASCII_GREATER;
+        case TOK_F2_EQUAL:      return out << UNI_EQUAL;
+        case TOK_F2_GREATER:    return out << UNI_GREATER;
         case TOK_F2_AND:        return out << UNI_AND;
         case TOK_F2_AND_B:      return out << UNI_AND << UNI_AND;
         case TOK_F2_OR:         return out << UNI_OR;
         case TOK_F2_OR_B:       return out << UNI_OR << UNI_OR;
         case TOK_F2_INDEX:      return out << UNI_SQUISH_Quad;
-        case TOK_F2_LEQ:        return out << UNI_LESS_OR_EQUAL;
-        case TOK_F2_MEQ:        return out << UNI_MORE_OR_EQUAL;
-        case TOK_F2_UNEQ:       return out << UNI_NOT_EQUAL;
+        case TOK_F2_LEQU:       return out << UNI_LESS_OR_EQUAL;
+        case TOK_F2_MEQU:       return out << UNI_MORE_OR_EQUAL;
+        case TOK_F2_UNEQU:      return out << UNI_NOT_EQUAL;
         case TOK_F2_NOR:        return out << UNI_NOR;
         case TOK_F2_NOR_B:      return out << UNI_NOR << UNI_NOR;
         case TOK_F2_NAND:       return out << UNI_NAND;
         case TOK_F2_NAND_B:     return out << UNI_NAND << UNI_NAND;
 
-        case TOK_F12_BINOM:     return out << UNI_ASCII_EXCLAM;
+        case TOK_F12_BINOM:     return out << UNI_EXCLAM;
         case TOK_F12_CIRCLE:    return out << UNI_CIRCLE;
-        case TOK_F12_COMMA:     return out << UNI_ASCII_COMMA;
+        case TOK_F12_COMMA:     return out << UNI_COMMA;
         case TOK_F12_COMMA1:    return out << UNI_COMMA_BAR;
         case TOK_F12_DECODE:    return out << UNI_UP_TACK;
         case TOK_F12_DIVIDE:    return out << UNI_DIVIDE;
@@ -329,15 +347,15 @@ Token::print_function(ostream & out) const
         case TOK_F12_FORMAT:    return out << UNI_FORMAT;
         case TOK_F12_INDEX_OF:  return out << UNI_IOTA;
         case TOK_F12_LOGA:      return out << UNI_LOGARITHM;
-        case TOK_F12_MINUS:     return out << UNI_ASCII_MINUS;
+        case TOK_F12_MINUS:     return out << UNI_MINUS;
         case TOK_F12_PARTITION: return out << UNI_SUBSET;
         case TOK_F12_PICK:      return out << UNI_SUPERSET;
-        case TOK_F12_PLUS:      return out << UNI_ASCII_PLUS;
+        case TOK_F12_PLUS:      return out << UNI_PLUS;
         case TOK_F12_POWER:     return out << UNI_STAR_OPERATOR;
         case TOK_F12_RHO:       return out << UNI_RHO;
         case TOK_F12_RND_DN:    return out << UNI_LEFT_FLOOR;
         case TOK_F12_RND_UP:    return out << UNI_LEFT_CEILING;
-        case TOK_F12_ROLL:      return out << UNI_ASCII_QUESTION;
+        case TOK_F12_ROLL:      return out << UNI_QUESTION;
         case TOK_F12_ROTATE:    return out << UNI_CIRCLE_STILE;
         case TOK_F12_ROTATE1:   return out << UNI_CIRCLE_BAR;
         case TOK_F12_SORT_ASC:  return out << UNI_SORT_ASCENDING;
@@ -384,7 +402,7 @@ Token::print_function(ostream & out) const
       }
    return out <<  ", }";
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 ostream &
 Token::print_value(ostream & out) const
 {
@@ -419,7 +437,7 @@ Token::print_value(ostream & out) const
                       }
 
                    const PrintContext pctx(PR_APL, 2, 80);
-                   out << UCS_string(v->get_ravel(e)
+                   out << UCS_string(v->get_cravel(e)
                                        .character_representation(pctx), 0, 80)
                        << " ";
                  }
@@ -460,7 +478,7 @@ Token::print_value(ostream & out) const
 
    return out <<  ", }";
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 void
 Token::show_trace(ostream & out, const UCS_string & fun_name, 
                   Function_Line line) const
@@ -511,7 +529,7 @@ const Value & val = *get_apl_val();
    else if (val.get_rank() == 1)   // vector
       {
         if (val.element_count() == 0 &&   // empty vector
-            val.get_ravel(0).is_simple_cell())
+            val.get_cfirst().is_simple_cell())
            {
              out << endl;
              return;
@@ -525,27 +543,26 @@ const Value & val = *get_apl_val();
       }
 
 PrintBuffer pb(val, pctx, 0);
-const UCS_string indent(fn.size(), UNI_ASCII_SPACE);
-   loop(l, pb.get_height())
+const UCS_string indent(fn.size(), UNI_SPACE);
+   loop(l, pb.get_row_count())
       {
         if (l)   out << indent;
         out << pb.get_line(l).no_pad() << endl;
       }
 
-   if (pb.get_height() == 0)   out << endl;
+   if (pb.get_row_count() == 0)   out << endl;
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 ostream &
 Token::print_quad(ostream & out) const
 {
    return out << UNI_Quad_Quad << get_Id();
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 UCS_string
 Token::canonical(PrintStyle style) const
 {
 UCS_string ucs;
-
    switch(get_Class())
       {
         case TC_ASSIGN:
@@ -558,31 +575,32 @@ UCS_string ucs;
 
         case TC_L_BRACK:
              if (get_tag() == TOK_L_BRACK)
-                ucs.append(UNI_ASCII_L_BRACK);
+                ucs.append(UNI_L_BRACK);
              else if (get_tag() == TOK_SEMICOL)
-                ucs.append(UNI_ASCII_SEMICOLON);
+                ucs.append(UNI_SEMICOLON);
              else
                 Assert(0);
-                break;
+             break;
 
         case TC_R_BRACK:
-             ucs.append(UNI_ASCII_R_BRACK);
+             ucs.append(UNI_R_BRACK);
              break;
 
         case TC_END:
+        case TC_DIAMOND:
              ucs.append(UNI_DIAMOND);
              break;
 
         case TC_RETURN:                                                  break;
         case TC_LINE:
-             ucs.append(UNI_ASCII_LF);
+             ucs.append(UNI_LF);
              break;
 
         case TC_VALUE:
              {
                PrintContext pctx(style, DEFAULT_Quad_PP, DEFAULT_Quad_PW);
                PrintBuffer pbuf(*get_apl_val(), pctx, 0);
-               if (pbuf.get_height() == 0)   return ucs;
+               if (pbuf.get_row_count() == 0)   return ucs;
                return pbuf.l1();
              }
 
@@ -590,7 +608,6 @@ UCS_string ucs;
              ucs.append(get_sym_ptr()->get_name());
              break;
 
-
         case TC_R_PARENT:
         case TC_L_PARENT:
         case TC_R_CURLY:
@@ -604,50 +621,69 @@ UCS_string ucs;
              if (get_Id() == ID_No_ID)   return get_function()->get_name();
              return ID::get_name_UCS(get_Id());
 
+
+        case TC_INDEX:
+             if (get_tag() == TOK_AXIS)
+                {
+                  UCS_string ret;
+                  ret << "[";
+
+                  // caution: get_apl_val() may be 0 (for index []).
+                  //
+                  if (const Value * axis = get_apl_val().get())
+                     {
+                      ret << ShapeItem(axis->get_cfirst().get_int_value());
+                     }
+                  ret << "]";
+                  return ret;
+                }
+
         default:
              CERR << "Token: " << HEX4(tag) << " " << *this
                   << " at " << LOC << endl;
              Q1(get_Class())
-             Backtrace::show(__FILE__, __LINE__);
+             BACKTRACE
              FIXME;
       }
 
    return ucs;
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 UCS_string
 Token::tag_name() const
 {
    switch(get_tag())
       {
-#define TD(tag, _tc, _tv, _id) case tag: return UCS_string( #tag );
+#define TD(tag, _tc, _tv, _id) case tag: return UCS_ASCII_string(#tag);
 #include "Token.def"
+        case TOK_NONE:   return UCS_ASCII_string("???Bad-Tag???");
       }
 
 char cc[40];
-   snprintf(cc, sizeof(cc), "0x%X", get_tag());
-UCS_string ucs(cc);
+   SPRINTF(cc, "0x%X", get_tag());
+const UTF8_string utf(cc);
+UCS_string ucs(utf);
    return ucs;
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 int
 Token::error_info(UCS_string & ucs) const
 {
-UCS_string can = canonical(PR_APL_FUN).remove_pad();
+UCS_string canon = canonical(PR_APL_FUN).remove_pad();
 
 const Unicode c1 = ucs.back();
-const Unicode c2 = can.size() ? can[0] : Invalid_Unicode;
+const Unicode c2 = canon.size() ? canon[0] : Invalid_Unicode;
 
    // conditions when we don't need a space
    //
 bool need_space = ! (Avec::no_space_after(c1) || Avec::no_space_before(c2));
 
-   if (need_space)   ucs.append(UNI_ASCII_SPACE);
+   if (need_space)   ucs.append(UNI_SPACE);
 
-   ucs.append(can);
-   return need_space ? -can.size() : can.size();
+   ucs.append(canon);
+   return need_space ? -canon.size() : canon.size();
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 const char * 
 Token::short_class_name(TokenClass cls)
 {
@@ -690,7 +726,7 @@ Token::short_class_name(TokenClass cls)
 
    return "???";
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 ostream &
 operator << (ostream & out, const Token_string & tos)
 {
@@ -700,7 +736,7 @@ operator << (ostream & out, const Token_string & tos)
    out << endl;
    return out;
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 const char *
 Token::class_name(TokenClass tc)
 {
@@ -745,85 +781,7 @@ Token::class_name(TokenClass tc)
 
    return "*** Obscure token class ***";
 }
-//-----------------------------------------------------------------------------
-inline void
-Token::copy_N(const Token & src)
-{
-   tag = src.tag;
-   switch(src.get_ValueType())
-      {
-        case TV_NONE:  value.int_vals[0]   = 0;
-                       value.int_vals[1]   = src.value.int_vals[1];     break;
-
-        case TV_CHAR:  value.char_val      = src.value.char_val;        break;
-        case TV_INT:   value.int_vals[0]   = src.value.int_vals[0];
-                       value.int_vals[1]   = src.value.int_vals[1];     break;
-        case TV_FLT:   value.float_vals[0] = src.value.float_vals[0];   break;
-        case TV_CPX:   value.float_vals[0] = src.value.float_vals[0];
-                       value.float_vals[1] = src.value.float_vals[1];   break;
-        case TV_SYM:   value.sym_ptr       = src.value.sym_ptr;         break;
-        case TV_LIN:   value.fun_line      = src.value.fun_line;        break;
-        case TV_VAL:   value._apl_val()    = src.value._apl_val();      break;
-        case TV_INDEX: value.index_val     = src.value.index_val;       break;
-        case TV_FUN:   value.function      = src.value.function;        break;
-        default:       FIXME;
-      }
-}
-//-----------------------------------------------------------------------------
-void
-Token::copy_1(const Token & src, const char * loc)
-{
-   clear(loc);
-   if (src.is_apl_val())
-      {
-        const Value * val = src.value.apl_val.get();
-        if (val)
-           {
-             ADD_EVENT(val, VHE_TokCopy1, src.value_use_count(), loc);
-           }
-        else
-           {
-             ADD_EVENT(val, VHE_TokCopy1, -1, loc);
-           }
-      }
-
-   copy_N(src);
-}
-//-----------------------------------------------------------------------------
-void
-Token::move_1(Token & src, const char * loc)
-{
-   clear(loc);
-   copy_N(src);
-
-   if (src.is_apl_val())
-      {
-        const Value * val = src.value.apl_val.get();
-        if (val)
-           {
-             ADD_EVENT(val, VHE_TokMove1, src.value_use_count() - 1, loc);
-             src.clear(loc);
-           }
-      }
-}
-//-----------------------------------------------------------------------------
-void
-Token::move_2(const Token & src, const char * loc)
-{
-   clear(loc);
-   copy_N(src);
-
-   if (src.is_apl_val())
-      {
-        const Value * val = src.value.apl_val.get();
-        if (val)
-           {
-             ADD_EVENT(val, VHE_TokMove2, src.value_use_count() - 1, loc);
-             const_cast<Token &>(src).clear(loc);
-           }
-      }
-}
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
 void
 Token_string::reverse_from_to(ShapeItem from, ShapeItem to)
 {
@@ -833,24 +791,44 @@ Token * t2 = &at(to);
    Assert(from <= to);
    Assert(to <= ShapeItem(size()));
 
-   while (t1 < t2)   t1++->Hswap(*t2--);
+   while (t1 < t2)   t1++->swap_token(*t2--);
 }
-//-----------------------------------------------------------------------------
-void
-Token_string::print(ostream & out, bool details) const
+//----------------------------------------------------------------------------
+ShapeItem
+Token_string::replace_segment(const Token_string & src, ShapeItem pos)
 {
-   loop(t, size())
+   loop(s, src.size())
        {
-         const Token & tok = at(t);
+         at(pos).clear(LOC);
+         new (&at(pos++)) Token(src[s], LOC);
+       }
+   return pos;
+}
+//----------------------------------------------------------------------------
+void
+Token_string::print(ostream & out, int details) const
+{
+const bool PC  = details    & 1;
+const bool VAL = details & 2;
+
+   loop(pc, size())
+       {
+         const Token & tok = at(pc);
+         if (PC)   out << "    [PC=" << setw(2) << pc << "] ";
          out << "`" << tok;
-         if (details)   switch(tok.get_ValueType())
+         if (VAL)
             {
-              case TV_FLT: out << " " << tok.get_flt_val();   break;
-              default:                                        break;
+              switch(tok.get_ValueType())
+                 {
+                   case TV_INT: out << ":" << tok.get_int_val();   break;
+                   case TV_FLT: out << ":" << tok.get_flt_val();   break;
+                   default:                                        break;
+                 }
             }
-         out << "  ";
+         if (PC)   out << endl;
+         else      out << "  ";
        }
 
    out << endl;
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
