@@ -18,6 +18,9 @@
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
+/** @file
+*/
+
 #include "emacs.hh"
 #include "network.hh"
 #include "../NativeFunction.hh"
@@ -90,7 +93,7 @@ Token eval_AB(Value_P A, Value_P B)
 
 Token eval_XB(Value_P X, Value_P B)
 {
-    const int function_number = X->get_ravel(0).get_near_int();
+    const int function_number = X->get_cravel(0).get_near_int();
 
     switch( function_number ) {
     case 0:
@@ -103,16 +106,17 @@ Token eval_XB(Value_P X, Value_P B)
             port = 0;
         }
         else {
-            port = B->get_ravel( 0 ).get_near_int();
+            port = B->get_cravel( 0 ).get_near_int();
         }
 
         try {
-            start_listener( port );
-        }
-        catch( InitProtocolError &error ) {
-            Workspace::more_error() = error.get_message().c_str();
-            DOMAIN_ERROR;
-        }
+              start_listener( port );
+            }
+        catch(InitProtocolError &error)
+           {
+             MORE_ERROR() << error.get_message().c_str();
+             DOMAIN_ERROR;
+           }
         return Token(TOK_APL_VALUE1, Str0(LOC));
     }
 
@@ -167,14 +171,14 @@ const UCS_string ucs_string_from_string( const std::string &string )
     return UCS_string( utf );
 }
 
-Value_P make_string_cell( const std::string &string, const char *loc )
+Value_P
+make_string_cell( const std::string &string, const char *loc )
 {
-    UCS_string s = ucs_string_from_string( string );
-    Shape shape( s.size() );
-    Value_P cell(shape, loc );
-    for( int i = 0 ; i < s.size() ; i++ ) {
-        new (cell->next_ravel()) CharCell( s[i] );
-    }
+UCS_string s = ucs_string_from_string( string );
+Shape shape(s.size());
+Value_P cell(shape, loc );
+
+    loop (i, s.size())   cell->next_ravel_Char(s[i]);
     cell->check_value( loc );
     return cell;
 }
