@@ -18,12 +18,14 @@
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
-#include "config.h"   // for HAVE_xxx macros
+/** @file
+*/
+
+#include <vector>
 
 #include "apl-sqlite.hh"
 #include "NativeFunction.hh"
 
-#include <vector>
 #include <map>
 #include <typeinfo>
 
@@ -52,22 +54,24 @@ extern "C" {
     void *get_function_mux( const char *function_name );
 }
 
-static void add_provider( Provider *provider )
+static void
+add_provider(Provider * provider)
 {
-    providers.insert( pair<const string, Provider *>( provider->get_name(), provider ) );
+   providers.insert(pair<const string, Provider *>
+                     (provider->get_provider_name(), provider) );
 }
 
 static void init_provider_map( void )
 {
 #ifdef HAVE_SQLITE3
     add_provider( new SqliteProvider() );
-#elif REALLY_WANT_SQLITE3
+#elif cfg_USER_WANTS_SQLITE3
 # warning "SQLite3 unavailable since ./configure could not detect it"
 #endif
 
-#ifdef USABLE_PostgreSQL
+#ifdef HAVE_POSTGRES
     add_provider( new PostgresProvider() );
-#elif REALLY_WANT_PostgreSQL
+#elif cfg_USER_WANTS_POSTGRES
 #  warning "PostgreSQL unavailable since ./configure could not detect it."
 # if HAVE_POSTGRESQL
 #  warning "The PostgreSQL library seems to be installed, but the header file(s) are missing"
@@ -105,17 +109,19 @@ static int find_free_connection( void )
 static Token open_database( Value_P A, Value_P B )
 {
     if( !A->is_apl_char_vector() ) {
-        Workspace::more_error() = "Illegal database name";
-        VALUE_ERROR;
-    }
-    string type = to_string( A->get_UCS_ravel() );
-    map<const string, Provider *>::iterator provider_iterator = providers.find( type );
-    if( provider_iterator == providers.end() ) {
-        stringstream out;
-        out << "Unknown database type: " << type;
-        Workspace::more_error() = out.str().c_str();
+        MORE_ERROR() << "Illegal database name";
         VALUE_ERROR;
     }
+    UTF8_string type(A->get_UCS_ravel());
+    map<const string, Provider *>::iterator provider_iterator =
+                                            providers.find(type.c_str());
+    if (provider_iterator == providers.end())
+       {
+         stringstream out;
+         out << "Unknown database type: " << type;
+         MORE_ERROR() << out.str().c_str();
+         VALUE_ERROR;
+       }
 
     int connection_index = find_free_connection();
     connections[connection_index] = provider_iterator->second->open_database( B );
@@ -125,7 +131,7 @@ static Token open_database( Value_P A, Value_P B )
 
 static void throw_illegal_db_id( void )
 {
-    Workspace::more_error() = "Illegal database id";
+    MORE_ERROR() << "Illegal database id";
     DOMAIN_ERROR;
 }
 
@@ -148,18 +154,18 @@ static Connection *value_to_db_id( Value_P value )
         throw_illegal_db_id();
     }
 
-    int db_id = value->get_ravel( 0 ).get_int_value();
+    int db_id = value->get_cravel( 0 ).get_int_value();
     return db_id_to_connection( db_id );
  }
 
 static Token close_database( Value_P B )
 {
     if( !B->is_int_scalar( ) ) {
-        Workspace::more_error() = "Close database command requires database id as argument";
+        MORE_ERROR() << "Close database command requires database id as argument";
         DOMAIN_ERROR;
     }
 
-    int db_id = B->get_ravel( 0 ).get_int_value();
+    int db_id = B->get_cravel( 0 ).get_int_value();
     if( db_id < 0 || db_id >= (int)connections.size() ) {
         throw_illegal_db_id();
     }
@@ -173,33 +179,40 @@ static Token close_database( Value_P B )
 
     return Token( TOK_APL_VALUE1, Str0( LOC ) );
 }
-
-static Value_P run_generic_one_query( ArgListBuilder *arg_list,
-                                      Value_P B, int start, int num_args,
-                                      bool ignore_result )
+//-----------------------------------------------------------------------------
+static Value_P
+run_generic_one_query(ArgListBuilder *arg_list, Value_P B,
+                      int start, int num_args, bool ignore_result)
 {
-    for( int i = 0 ; i < num_args ; i++ ) {
-        const Cell &cell = B->get_ravel( start + i );
-        if( cell.is_integer_cell() ) {
-            arg_list->append_long( cell.get_int_value(), i );
-        }
-        else if( cell.is_float_cell() ) {
-            arg_list->append_double( cell.get_real_value(), i );
-        }
-        else {
-            Value_P value = cell.to_value( LOC );
-            if( value->get_shape().get_volume() == 0 ) {
-                arg_list->append_null( i );
+   loop(i, num_args)
+       {
+         const Cell & cell_B = B->get_cravel(start + i);
+         if (cell_B.is_integer_cell())
+            {
+              arg_list->append_long(cell_B.get_int_value(), i);
             }
-            else if( value->is_char_string() ) {
-                arg_list->append_string( to_string( value->get_UCS_ravel() ), i );
-            }
-            else {
-                stringstream out;
-                out << "Illegal data type in argument " << i << " of arglist";
-                Workspace::more_error() = out.str().c_str();
-                VALUE_ERROR;
+        else if (cell_B.is_float_cell())
+            {
+              arg_list->append_double(cell_B.get_real_value(), i);
             }
+        else
+           {
+             Value_P value = cell_B.to_value(LOC);
+             if (value->get_shape().get_volume() == 0)
+                {
+                  arg_list->append_null(i);
+                }
+             else if (value->is_char_string())
+                {
+                  const UTF8_string utf(value->get_UCS_ravel());
+                  arg_list->append_string(utf.c_str(), i);
+                }
+            else
+                {
+                  MORE_ERROR() << "Illegal data type in argument "
+                               << i << " of arglist";
+                  DOMAIN_ERROR;
+                }
         }
     }
 
@@ -209,50 +222,57 @@ static Value_P run_generic_one_query( ArgListBuilder *arg_list,
 static Value_P run_generic( Connection *conn, Value_P A, Value_P B, bool query )
 {
     if( !A->is_char_string() ) {
-        Workspace::more_error() = "Illegal query argument type";
+        MORE_ERROR() << "Illegal query argument type";
         VALUE_ERROR;
     }
 
-    string statement = conn->replace_bind_args( to_string( A->get_UCS_ravel() ) );
-    ArgListBuilder *builder;
-    if( query ) {
-        builder = conn->make_prepared_query( statement );
-    }
-    else {
-        builder = conn->make_prepared_update( statement );
-    }
-    auto_ptr<ArgListBuilder> arg_list( builder );
-
-    const Shape &shape = B->get_shape();
-    if( shape.get_rank() == 0 || shape.get_rank() == 1 ) {
-        int num_args = shape.get_volume();
-        return run_generic_one_query( arg_list.get(), B, 0, num_args, false );
-    }
-    else if( shape.get_rank() == 2 ) {
-        int rows = shape.get_rows();
-        int cols = shape.get_cols();
-        if( rows == 0 ) {
-            return Idx0( LOC );
+     const UTF8_string A_utf(A->get_UCS_ravel());
+     string statement = conn->replace_bind_args(A_utf.c_str());
+     ArgListBuilder * builder;
+     if (query)
+        {
+          builder = conn->make_prepared_query(statement);
         }
-        else {
-            Assert_fatal( rows > 0 );
-            Value_P result;
-            for( int row = 0 ; row < rows ; row++ ) {
-                bool not_last = row < rows - 1;
-                result = run_generic_one_query( arg_list.get(), B, row * cols, cols, not_last );
-                if( not_last ) {
-                    arg_list->clear_args();
-                }
+    else {
+           builder = conn->make_prepared_update(statement);
+         }
+    unique_ptr<ArgListBuilder> arg_list(builder);
+
+    const Shape & shape = B->get_shape();
+    if (shape.get_rank() == 0 || shape.get_rank() == 1)
+       {
+         const int num_args = shape.get_volume();
+         return run_generic_one_query( arg_list.get(), B, 0, num_args, false );
+       }
+    else if (shape.get_rank() == 2)
+       {
+         const int rows = shape.get_rows();
+         const int cols = shape.get_cols();
+         if (rows == 0)
+            {
+              return Idx0(LOC);
             }
-            return result;
-        }
-    }
-    else {
-        Workspace::more_error() = "Bind params have illegal rank";
-        RANK_ERROR;
-    }
+        else
+            {
+              Assert_fatal(rows > 0);
+              Value_P result;
+              for (int row = 0 ; row < rows ; row++)
+                  {
+                    bool not_last = row < rows - 1;
+                    result = run_generic_one_query(arg_list.get(), B, row*cols,
+                                                   cols, not_last);
+                    if (not_last)   arg_list->clear_args();
+                  }
+               return result;
+            }
+       }
+    else
+       {
+         MORE_ERROR() << "Bind params have illegal rank > 2";
+         RANK_ERROR;
+       }
 }
-
+//-----------------------------------------------------------------------------
 static Token run_query( Connection *conn, Value_P A, Value_P B )
 {
     return Token( TOK_APL_VALUE1, run_generic( conn, A, B, true ) );
@@ -284,24 +304,28 @@ static Token run_transaction_rollback( Value_P B )
     return Token( TOK_APL_VALUE1, Idx0( LOC ) );
 }
 
-static Token show_tables( Value_P B )
+static Token
+show_tables( Value_P B )
 {
-    Connection *conn = value_to_db_id( B );
-    vector<string> tables;
-    conn->fill_tables( tables );
+Connection * conn = value_to_db_id(B);
+vector<string> tables;
+   conn->fill_tables( tables );
 
-    Value_P value;
-    if( tables.size() == 0 ) {
+Value_P value;
+   if (tables.size() == 0)
+      {
         value = Idx0( LOC );
-    }
-    else {
-        Shape shape( tables.size () );
+      }
+    else
+      {
+        Shape shape(tables.size ());
         value = Value_P( shape, LOC );
-        for( vector<string>::iterator i = tables.begin() ; i != tables.end() ; i++ ) {
-            new (value->next_ravel()) PointerCell( make_string_cell( *i, LOC ),
-               value.getref() );
-        }
-    }
+        for (vector<string>::iterator i = tables.begin();
+             i != tables.end(); i++ )
+            {
+              value->next_ravel_Pointer(make_string_cell(*i, LOC).get());
+            }
+      }
 
     value->check_value( LOC );
     return Token( TOK_APL_VALUE1, value );
@@ -313,12 +337,12 @@ static Token show_cols( Value_P A, Value_P B )
     vector<ColumnDescriptor> cols;
 
     if( !B->is_apl_char_vector() ) {
-        Workspace::more_error() = "Illegal table name";
+        MORE_ERROR() << "Illegal table name";
         VALUE_ERROR;
     }
 
-    string name = to_string( B->get_UCS_ravel() );
-    conn->fill_cols( name, cols );
+    const UTF8_string name(B->get_UCS_ravel());
+    conn->fill_cols(name.c_str(), cols);
 
     Value_P value;
     if( cols.size() == 0 ) {
@@ -328,7 +352,7 @@ static Token show_cols( Value_P A, Value_P B )
         Shape shape( cols.size(), 2 );
         value = Value_P( shape, LOC );
         for( vector<ColumnDescriptor>::iterator i = cols.begin() ; i != cols.end() ; i++ ) {
-            new (value->next_ravel()) PointerCell( make_string_cell( i->get_name(), LOC ) , value.getref());
+            value->next_ravel_Pointer(make_string_cell(i->get_name(), LOC).get());
 
             Value_P type;
             if( i->get_type().size() == 0 ) {
@@ -337,7 +361,7 @@ static Token show_cols( Value_P A, Value_P B )
             else {
                 type = make_string_cell( i->get_type(), LOC );
             }
-            new (value->next_ravel()) PointerCell( type, value.getref() );
+            value->next_ravel_Pointer(type.get());
         }
     }
 
@@ -374,46 +398,35 @@ Token eval_AB( Value_P A, Value_P B )
 
 Token eval_XB(Value_P X, Value_P B)
 {
-    const int function_number = X->get_ravel( 0 ).get_near_int( );
+const int function_number = X->get_cfirst().get_near_int( );
 
-    switch( function_number ) {
-    case 0:
-        return list_functions( CERR );
-
-    case 2:
-        return close_database( B );
-
-    case 5:
-        return run_transaction_begin( B );
-
-    case 6:
-        return run_transaction_commit( B );
-
-    case 7:
-        return run_transaction_rollback( B );
-
-    case 8:
-        return show_tables( B );
-
-    default:
-        Workspace::more_error() = "Illegal function number";
-        DOMAIN_ERROR;
-    }
+    switch(function_number)
+       {
+         case 0:  return list_functions(CERR);
+         case 2:  return close_database(B);
+         case 5:  return run_transaction_begin(B);
+         case 6:  return run_transaction_commit(B);
+         case 7:  return run_transaction_rollback(B);
+         case 8:  return show_tables(B);
+         default: MORE_ERROR() << "Illegal function number";
+                  DOMAIN_ERROR;
+       }
 }
 
-static Connection *param_to_db( Value_P X )
+static Connection *
+param_to_db( Value_P X)
 {
     const Shape &shape = X->get_shape();
     if( shape.get_volume() != 2 ) {
-        Workspace::more_error() = "Database id missing from axis parameter";
+        MORE_ERROR() << "Database id missing from axis parameter";
         RANK_ERROR;
     }
-    return db_id_to_connection( X->get_ravel( 1 ).get_near_int( ) );
+    return db_id_to_connection( X->get_cravel( 1 ).get_near_int( ) );
 }
 
 Token eval_AXB(const Value_P A, const Value_P X, const Value_P B)
 {
-    const int function_number = X->get_ravel( 0 ).get_near_int( );
+    const int function_number = X->get_cravel( 0 ).get_near_int( );
 
     switch( function_number ) {
     case 0:
@@ -432,38 +445,27 @@ Token eval_AXB(const Value_P A, const Value_P X, const Value_P B)
         return show_cols( A, B );
 
     default:
-        Workspace::more_error() = "Illegal function number";
+        MORE_ERROR() << "Illegal function number";
         DOMAIN_ERROR;
     }
 }
 
-void *get_function_mux( const char *function_name )
+void *
+get_function_mux( const char *function_name )
 {
-    if( strcmp( function_name, "get_signature" ) == 0 ) return (void *)&get_signature;
-    if( strcmp( function_name, "eval_B" ) == 0 )        return (void *)&eval_B;
-    if( strcmp( function_name, "eval_AB" ) == 0 )       return (void *)&eval_AB;
-    if( strcmp( function_name, "eval_XB" ) == 0 )       return (void *)&eval_XB;
-    if( strcmp( function_name, "eval_AXB" ) == 0 )      return (void *)&eval_AXB;
-    if( strcmp( function_name, "close_fun" ) == 0 )     return (void *)&close_fun;
-    return 0;
+   if (strcmp(function_name, "get_signature"))   return (void *)&get_signature;
+   if (!strcmp(function_name, "eval_B"))         return (void *)&eval_B;
+   if (!strcmp(function_name, "eval_AB"))        return (void *)&eval_AB;
+   if (!strcmp(function_name, "eval_XB"))        return (void *)&eval_XB;
+   if (!strcmp(function_name, "eval_AXB"))       return (void *)&eval_AXB;
+   if (!strcmp( function_name, "close_fun"))     return (void *)&close_fun;
+   return 0;
 }
 
-static const UCS_string ucs_string_from_string( const std::string &string )
+Value_P
+make_string_cell(const std::string &str, const char * loc)
 {
-    size_t length = string.size();
-    const char *buf = string.c_str();
-    UTF8_string utf( (const UTF8 *)buf, length );
-    return UCS_string( utf );
-}
-
-Value_P make_string_cell( const std::string &string, const char *loc )
-{
-    UCS_string s = ucs_string_from_string( string );
-    Shape shape( s.size() );
-    Value_P cell( shape, loc );
-    for( int i = 0 ; i < s.size() ; i++ ) {
-        new (cell->next_ravel()) CharCell( s[i] );
-    }
-    cell->check_value( loc );
-    return cell;
+const UCS_string ucs(UTF8_string(str.c_str()));
+Value_P Z(ucs, loc);
+   return Z;
 }
